#
# Copyright 2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Supports a hybrid Unicode string that can also have a list of alternate
strings in the strings attribute.
"""

from __future__ import annotations


class multistring(str):
    def __new__(cls, string: list[str] | str | None = None):
        if string is None:
            string = [""]
        elif isinstance(string, str):
            string = [string]
        if not isinstance(string, list) or any(
            not isinstance(value, str) for value in string
        ):
            raise TypeError("multistring can only contain strings or list of strings")
        if not string:
            raise ValueError("multistring must contain at least one string")

        newstring = str.__new__(cls, string[0])
        newstring.extra_strings = string[1:]
        return newstring

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        if not hasattr(self, "extra_strings"):
            self.extra_strings = []

    @property
    def strings(self) -> list[str]:
        return [self, *self.extra_strings]

    def __hash__(self) -> int:
        return super().__hash__()

    def __bool__(self) -> bool:
        return len(self) > 0 or any(self.extra_strings)

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __eq__(self, other) -> bool:
        if isinstance(other, multistring):
            return super().__eq__(other) and self.extra_strings == other.extra_strings
        if isinstance(other, str):
            return super().__eq__(other)
        if isinstance(other, list):
            return self.strings == other
        return super().__eq__(other)

    def __repr__(self) -> str:
        strings = [str(self), *self.extra_strings]
        return f"multistring({strings!r})"

    def replace(self, old: str, new: str, count: int = -1) -> multistring:  # ty:ignore[invalid-method-override]
        newstr = multistring(super().replace(old, new, count))
        newstr.extra_strings.extend(
            s.replace(old, new, count) for s in self.extra_strings
        )
        return newstr



#
# Copyright 2002-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import fnmatch
import logging
import optparse
import os.path
import re
import sys
import traceback
from io import BytesIO
from types import TracebackType
from typing import Any

from translate import __version__
from translate.misc import progressbar


class ProgressBar:
    progress_types: dict[str, type[progressbar.ProgressBase]] = {
        "dots": progressbar.DotsProgressBar,
        "none": progressbar.NoProgressBar,
        "bar": progressbar.HashProgressBar,
        "names": progressbar.MessageProgressBar,
        "verbose": progressbar.VerboseProgressBar,
    }

    def __init__(self, progress_type, allfiles) -> None:
        """Set up a progress bar appropriate to the progress_type and files."""
        if progress_type in {"bar", "verbose"}:
            file_count = len(allfiles)
            self._progressbar = self.progress_types[progress_type](0, file_count)
            logger = logging.getLogger(os.path.basename(sys.argv[0])).getChild(
                "progress"
            )
            logger.setLevel(logging.INFO)
            logger.propagate = False
            handler = logging.StreamHandler()
            handler.setLevel(logging.INFO)
            handler.setFormatter(logging.Formatter())
            logger.addHandler(handler)
            logger.info("processing %d files...", file_count)
        else:
            self._progressbar = self.progress_types[progress_type]()

    def report_progress(self, filename, success) -> None:
        """Show that we are progressing..."""
        self._progressbar.amount += 1
        self._progressbar.show(filename)


class RecursiveOption(optparse.Option):
    """Option type with extra usage metadata used by RecursiveOptionParser."""

    optionalswitch: bool = False
    required: bool = False


class ManPageOption(RecursiveOption):
    ACTIONS = (*optparse.Option.ACTIONS, "manpage")

    def take_action(self, action, dest, opt, value, values, parser):
        """take_action that can handle manpage as well as standard actions."""
        if action == "manpage":
            parser.print_manpage()
            sys.exit(0)
        return super().take_action(action, dest, opt, value, values, parser)


class ManHelpFormatter(optparse.HelpFormatter):
    def __init__(
        self, indent_increment=0, max_help_position=0, width=80, short_first=1
    ) -> None:
        super().__init__(indent_increment, max_help_position, width, short_first)

    def format_option_strings(self, option):
        """Return a comma-separated list of option strings & metavariables."""
        if option.takes_value():
            metavar = option.metavar or option.dest.upper()
            metavar = f"\\fI{metavar}\\fP"
            short_opts = [sopt + metavar for sopt in option._short_opts]
            long_opts = [f"{lopt}\\fR=\\fP{metavar}" for lopt in option._long_opts]
        else:
            short_opts = option._short_opts
            long_opts = option._long_opts

        opts = short_opts + long_opts if self.short_first else long_opts + short_opts

        return "\\fB{}\\fP".format("\\fR, \\fP".join(opts))


class StdoutWrapper:
    def __init__(self) -> None:
        self.out = sys.stdout

    def __getattr__(self, name):
        return getattr(self.out, name)

    def write(self, content) -> None:
        if isinstance(content, bytes):
            try:
                self.out.write(content.decode("utf-8"))
            except UnicodeDecodeError:
                self.out.write("Unable to write binary content to the terminal")
        else:
            self.out.write(content)


class RecursiveOptionParser(optparse.OptionParser):
    """A specialized Option Parser for recursing through directories."""

    def __init__(
        self,
        formats: dict,
        usetemplates: bool = False,
        allowmissingtemplate: bool = False,
        description: str | None = None,
    ) -> None:
        """
        Construct the specialized Option Parser.

        :param formats: See :meth:`~.RecursiveOptionParser.setformats`
        for an explanation of the formats parameter.
        """
        super().__init__(version=f"%prog {__version__.sver}", description=description)
        self.setmanpageoption()
        self.setprogressoptions()
        self.seterrorleveloptions()
        self.setformats(formats, usetemplates)
        self.passthrough = []
        self.allowmissingtemplate = allowmissingtemplate
        logging.basicConfig(format="%(name)s: %(levelname)s: %(message)s")

    def get_prog_name(self):
        return os.path.basename(sys.argv[0])

    def setmanpageoption(self) -> None:
        """
        Creates a manpage option that allows the optionparser to generate a
        manpage.
        """
        manpageoption = ManPageOption(
            None,
            "--manpage",
            dest="manpage",
            default=False,
            action="manpage",
            help="output a manpage based on the help",
        )
        self.define_option(manpageoption)

    def format_manpage(self):
        """Returns a formatted manpage."""
        result = []
        prog = self.get_prog_name()

        def formatprog(x):
            return x.replace("%prog", prog)

        def formatToolkit(x):
            return x.replace("%prog", "Translate Toolkit")

        result.extend(
            (
                '.\\" Autogenerated manpage\n',
                f'.TH {prog} 1 "{formatToolkit(self.version)}" "" "{formatToolkit(self.version)}"\n',
                ".SH NAME\n",
                "{} \\- {}\n".format(
                    self.get_prog_name(),
                    self.description.split("\n\n")[0],  # ty:ignore[unresolved-attribute]
                ),
                ".SH SYNOPSIS\n",
                ".PP\n",
            )
        )
        usage = "\\fB%prog "
        usage += " ".join(self.getusageman(option) for option in self.option_list)
        usage += "\\fP"
        result.append(f"{formatprog(usage)}\n")
        description_lines = self.description.split("\n\n")[1:]  # ty:ignore[unresolved-attribute]
        if description_lines:
            result.extend(
                (
                    ".SH DESCRIPTION\n",
                    "\n\n".join(
                        re.sub(r"\.\. note::", "Note:", line)
                        for line in description_lines
                    ),
                )
            )
        result.append(".SH OPTIONS\n")
        ManHelpFormatter().store_option_strings(self)
        result.append(".PP\n")
        for option in self.option_list:
            result.extend(
                (
                    ".TP\n",
                    "{}\n".format(str(option).replace("-", "\\-")),
                    "{}\n".format(option.help.replace("-", "\\-")),  # ty:ignore[unresolved-attribute]
                )
            )
        return "".join(result)

    def print_manpage(self, file=None) -> None:
        """Outputs a manpage for the program using the help information."""
        if file is None:
            file = sys.stdout
        file.write(self.format_manpage())

    def set_usage(self, usage=None) -> None:
        """
        Sets the usage string - if usage not given, uses getusagestring for
        each option.
        """
        if usage is None:
            self.usage = f"%prog {' '.join(self.getusagestring(option) for option in self.option_list)}"
        else:
            super().set_usage(usage)

    def warning(
        self,
        msg,
        options=None,
        exc_info: tuple[type[BaseException], BaseException, TracebackType]
        | tuple[None, None, None]
        | None = None,
    ) -> None:
        """Print a warning message incorporating 'msg' to stderr."""
        if options:
            if options.errorlevel == "traceback":
                assert exc_info is not None
                assert exc_info[0] is not None
                errorinfo = "\n".join(
                    traceback.format_exception(exc_info[0], exc_info[1], exc_info[2])
                )
            elif options.errorlevel == "exception":
                assert exc_info is not None
                assert exc_info[0] is not None
                errorinfo = "\n".join(
                    traceback.format_exception_only(exc_info[0], exc_info[1])
                )
            elif options.errorlevel == "message":
                assert exc_info is not None
                assert exc_info[1] is not None
                errorinfo = str(exc_info[1])
            else:
                errorinfo = ""
            if errorinfo:
                msg += f": {errorinfo}"
        logging.getLogger(self.get_prog_name()).warning(msg)

    @staticmethod
    def getusagestring(option):
        """Returns the usage string for the given option."""
        optionstring = "|".join(option._short_opts + option._long_opts)
        if getattr(option, "optionalswitch", False):
            optionstring = f"[{optionstring}]"
        if option.metavar:
            optionstring += f" {option.metavar}"
        if getattr(option, "required", False):
            return optionstring
        return f"[{optionstring}]"

    @staticmethod
    def getusageman(option):
        """Returns the usage string for the given option."""
        optionstring = "\\fR|\\fP".join(option._short_opts + option._long_opts)
        if getattr(option, "optionalswitch", False):
            optionstring = f"\\fR[\\fP{optionstring}\\fR]\\fP"
        if option.metavar:
            optionstring += f" \\fI{option.metavar}\\fP"
        if getattr(option, "required", False):
            return optionstring
        return f"\\fR[\\fP{optionstring}\\fR]\\fP"

    def define_option(self, option) -> None:
        """
        Defines the given option, replacing an existing one of the same
        short name if necessary...
        """
        for short_opt in option._short_opts:
            if self.has_option(short_opt):
                self.remove_option(short_opt)
        for long_opt in option._long_opts:
            if self.has_option(long_opt):
                self.remove_option(long_opt)
        self.add_option(option)

    def setformats(
        self, formats: dict | list[tuple[Any, Any]], usetemplates: bool
    ) -> None:
        """
        Sets the format options using the given format dictionary.

        :param formats: The dictionary *keys* should be:

                        - Single strings (or 1-tuples) containing an
                          input format (if not *usetemplates*)
                        - Tuples containing an input format and
                          template format (if *usetemplates*)
                        - Formats can be *None* to indicate what to do
                          with standard input

                        The dictionary *values* should be tuples of
                        outputformat (string) and processor method.
        """
        self.inputformats = []
        outputformats = []
        templateformats = []
        self.outputoptions = {}
        self.usetemplates = usetemplates
        format_items: list[tuple[Any, Any]] = (
            list(formats.items()) if isinstance(formats, dict) else formats
        )
        for formatgroup, outputoptions in format_items:
            if isinstance(formatgroup, str) or formatgroup is None:
                formatgroup = (formatgroup,)
            if not isinstance(formatgroup, tuple):
                raise TypeError("formatgroups must be tuples or None/str/unicode")
            if len(formatgroup) < 1 or len(formatgroup) > 2:
                raise ValueError("formatgroups must be tuples of length 1 or 2")
            if len(formatgroup) == 1:
                formatgroup += (None,)
            inputformat, templateformat = formatgroup
            if not isinstance(outputoptions, tuple) or len(outputoptions) != 2:
                raise ValueError("output options must be tuples of length 2")
            outputformat, processor = outputoptions
            if inputformat not in self.inputformats:
                self.inputformats.append(inputformat)
            if outputformat not in outputformats:
                outputformats.append(outputformat)
            if templateformat not in templateformats:
                templateformats.append(templateformat)
            self.outputoptions[inputformat, templateformat] = (
                outputformat,
                processor,
            )
        inputformathelp = self.getformathelp(self.inputformats)
        inputoption = RecursiveOption(
            "-i",
            "--input",
            dest="input",
            action="append",
            default=None,
            metavar="INPUT",
            help=f"read from INPUT in {inputformathelp}",
        )
        inputoption.optionalswitch = True
        inputoption.required = True
        self.define_option(inputoption)
        excludeoption = RecursiveOption(
            "-x",
            "--exclude",
            dest="exclude",
            action="append",
            type="string",
            metavar="EXCLUDE",
            default=["CVS", ".svn", "_darcs", ".git", ".hg", ".bzr"],
            help="exclude names matching EXCLUDE from input paths",
        )
        self.define_option(excludeoption)
        outputformathelp = self.getformathelp(outputformats)
        outputoption = RecursiveOption(
            "-o",
            "--output",
            dest="output",
            default=None,
            metavar="OUTPUT",
            help=f"write to OUTPUT in {outputformathelp}",
        )
        outputoption.optionalswitch = True
        outputoption.required = True
        self.define_option(outputoption)
        if self.usetemplates:
            self.templateformats = templateformats
            templateformathelp = self.getformathelp(self.templateformats)
            templateoption = RecursiveOption(
                "-t",
                "--template",
                dest="template",
                default=None,
                metavar="TEMPLATE",
                help=f"read from TEMPLATE in {templateformathelp}",
            )
            self.define_option(templateoption)

    def setprogressoptions(self) -> None:
        """Sets the progress options."""
        progressoption = RecursiveOption(
            None,
            "--progress",
            dest="progress",
            default="bar",
            choices=list(ProgressBar.progress_types.keys()),
            metavar="PROGRESS",
            help=f"show progress as: {', '.join(ProgressBar.progress_types)}",
        )
        self.define_option(progressoption)

    def seterrorleveloptions(self) -> None:
        """Sets the errorlevel options."""
        self.errorleveltypes = ["none", "message", "exception", "traceback"]
        errorleveloption = RecursiveOption(
            None,
            "--errorlevel",
            dest="errorlevel",
            default="message",
            choices=self.errorleveltypes,
            metavar="ERRORLEVEL",
            help=f"show errorlevel as: {', '.join(self.errorleveltypes)}",
        )
        self.define_option(errorleveloption)

    @staticmethod
    def getformathelp(formats) -> str:
        """Make a nice help string for describing formats..."""
        formats = sorted(f for f in formats if f is not None)
        if len(formats) == 0:
            return ""
        if len(formats) == 1:
            return f"{', '.join(formats)} format"
        return f"{', '.join(formats)} formats"

    @staticmethod
    def isrecursive(fileoption, filepurpose="input"):
        """Checks if fileoption is a recursive file."""
        if fileoption is None:
            return False
        if isinstance(fileoption, list):
            return True
        return os.path.isdir(fileoption)

    def parse_args(self, args=None, values=None):
        """
        Parses the command line options, handling implicit input/output
        args.
        """
        (options, args) = super().parse_args(args, values)
        # some intelligent as to what reasonable people might give on the
        # command line
        if args and not options.input:
            if len(args) > 1:
                options.input = args[:-1]
                args = args[-1:]
            else:
                options.input = args[0]
                args = []
        if args and not options.output:
            options.output = args[-1]
            args = args[:-1]
        if args:
            self.error(
                "You have used an invalid combination of --input, --output and freestanding args"
            )
        if isinstance(options.input, list) and len(options.input) == 1:
            options.input = options.input[0]
        if options.input is None:
            self.error(
                "You need to give an inputfile or use - for stdin ; use --help for full usage instructions"
            )
        elif options.input == "-":
            options.input = None
        return (options, args)

    def getpassthroughoptions(self, options):
        """Get the options required to pass to the filtermethod..."""
        passthroughoptions = {}
        for optionname in dir(options):
            if optionname in self.passthrough:
                passthroughoptions[optionname] = getattr(options, optionname)
        return passthroughoptions

    def getoutputoptions(self, options, inputpath, templatepath):
        """Works out which output format and processor method to use..."""
        if inputpath:
            _inputbase, inputext = self.splitinputext(inputpath)
        else:
            inputext = None
        if templatepath:
            _templatebase, templateext = self.splittemplateext(templatepath)
        else:
            templateext = None
        if (inputext, templateext) in self.outputoptions:
            return self.outputoptions[inputext, templateext]
        if (inputext, "*") in self.outputoptions:
            outputformat, fileprocessor = self.outputoptions[inputext, "*"]
        elif ("*", templateext) in self.outputoptions:
            outputformat, fileprocessor = self.outputoptions["*", templateext]
        elif ("*", "*") in self.outputoptions:
            outputformat, fileprocessor = self.outputoptions["*", "*"]
        elif (inputext, None) in self.outputoptions:
            return self.outputoptions[inputext, None]
        elif (None, templateext) in self.outputoptions:
            return self.outputoptions[None, templateext]
        elif ("*", None) in self.outputoptions:
            outputformat, fileprocessor = self.outputoptions["*", None]
        elif (None, "*") in self.outputoptions:
            outputformat, fileprocessor = self.outputoptions[None, "*"]
        else:
            if self.usetemplates:
                if inputext is None:
                    raise ValueError(
                        "don't know what to do with input format (no file extension), no template file"
                    )
                if templateext is None:
                    raise ValueError(
                        f"don't know what to do with input format {os.extsep + inputext}, no template file"
                    )
                raise ValueError(
                    f"don't know what to do with input format {os.extsep + inputext}, template format {os.extsep + templateext}"
                )
            if inputext is None:
                raise ValueError(
                    "don't know what to do with input format (no file extension)"
                )
            raise ValueError(
                f"don't know what to do with input format {os.extsep + inputext}"
            )
        if outputformat == "*":
            if inputext:
                outputformat = inputext
            elif templateext:
                outputformat = templateext
            elif ("*", "*") in self.outputoptions:
                outputformat = None
            else:
                if self.usetemplates:
                    raise ValueError(
                        "don't know what to do with input format (no file extension), no template file"
                    )
                raise ValueError(
                    "don't know what to do with input format (no file extension)"
                )
        return outputformat, fileprocessor

    @staticmethod
    def getfullinputpath(options, inputpath):
        """Gets the full path to an input file."""
        if options.input:
            return os.path.join(options.input, inputpath)
        return inputpath

    @staticmethod
    def normalizerelativepath(pathname, allow_empty: bool = False):
        """Normalizes a recursive relative path and rejects path traversal."""
        if pathname is None:
            return None
        normalized = os.path.normpath(pathname)
        drive, _tail = os.path.splitdrive(normalized)
        if drive or os.path.isabs(normalized):
            raise ValueError(
                f"unsafe recursive path {pathname!r}: absolute paths are not allowed"
            )
        if normalized in {"", "."}:
            if allow_empty:
                return ""
            raise ValueError(
                f"unsafe recursive path {pathname!r}: empty output paths are not allowed"
            )
        if normalized == os.pardir or normalized.startswith(os.pardir + os.sep):
            raise ValueError(
                f"unsafe recursive path {pathname!r}: path traversal outside the output directory is not allowed"
            )
        return normalized

    @staticmethod
    def getfulloutputpath(options, outputpath):
        """Gets the full path to an output file."""
        if options.recursiveoutput and options.output and outputpath is not None:
            outputpath = RecursiveOptionParser.normalizerelativepath(outputpath)
            return os.path.join(options.output, outputpath)
        return outputpath

    def getfulltemplatepath(self, options, templatepath):
        """Gets the full path to a template file."""
        if not options.recursivetemplate:
            return templatepath
        if templatepath is not None and self.usetemplates and options.template:
            return os.path.join(options.template, templatepath)
        return None

    def run(self) -> None:
        """
        Parses the arguments, and runs recursiveprocess with the resulting
        options...
        """
        (options, _args) = self.parse_args()
        self.recursiveprocess(options)

    def recursiveprocess(self, options) -> None:
        """Recurse through directories and process files."""
        if self.isrecursive(options.input, "input") and getattr(
            options, "allowrecursiveinput", True
        ):
            self.ensurerecursiveoutputdirexists(options)
            if isinstance(options.input, list):
                inputfiles = self.recurseinputfilelist(options)
            else:
                inputfiles = self.recurseinputfiles(options)
        elif options.input:
            inputfiles = [os.path.basename(options.input)]
            options.input = os.path.dirname(options.input)
        else:
            inputfiles = [options.input]
        options.recursiveoutput = self.isrecursive(
            options.output, "output"
        ) and getattr(options, "allowrecursiveoutput", True)
        options.recursivetemplate = (
            self.usetemplates
            and self.isrecursive(options.template, "template")
            and getattr(options, "allowrecursivetemplate", True)
        )
        # sort the input files to preserve the order between runs as much as possible.
        # this makes for more merge-friendly content in single-output-file mode.
        inputfiles.sort()
        progress_bar = ProgressBar(options.progress, inputfiles)
        for inputpath in inputfiles:
            try:
                templatepath = self.gettemplatename(options, inputpath)
                # If we have a recursive template, but the template doesn't
                # have this input file, let's drop it.
                if (
                    options.recursivetemplate
                    and templatepath is None
                    and not self.allowmissingtemplate
                ):
                    self.warning(
                        f"No template at {templatepath}. Skipping {inputpath}."
                    )
                    continue
                outputformat, fileprocessor = self.getoutputoptions(
                    options, inputpath, templatepath
                )
                fullinputpath = self.getfullinputpath(options, inputpath)
                fulltemplatepath = self.getfulltemplatepath(options, templatepath)
                outputpath = self.getoutputname(options, inputpath, outputformat)
                fulloutputpath = self.getfulloutputpath(options, outputpath)
                if options.recursiveoutput and outputpath:
                    self.checkoutputsubdir(options, os.path.dirname(outputpath))
            except Exception:
                self.warning(
                    f"Couldn't handle input file {inputpath}", options, sys.exc_info()
                )
                continue
            try:
                success = self.processfile(
                    fileprocessor,
                    options,
                    fullinputpath,
                    fulloutputpath,
                    fulltemplatepath,
                )
            except Exception:
                self.warning(
                    f"Error processing: input {fullinputpath}, output {fulloutputpath}, template {fulltemplatepath}",
                    options,
                    sys.exc_info(),
                )
                success = False
            progress_bar.report_progress(inputpath, success)

    def ensurerecursiveoutputdirexists(self, options) -> None:
        if not self.isrecursive(options.output, "output"):
            if not options.output:
                self.error("No output directory given")
            try:
                self.warning("Output directory does not exist. Attempting to create")
                os.mkdir(options.output)
            except OSError:
                self.error("Output directory does not exist, attempt to create failed")

    @staticmethod
    def openinputfile(options, fullinputpath):
        """Opens the input file."""
        if fullinputpath is None:
            return sys.stdin
        return open(fullinputpath, "rb")

    @staticmethod
    def openoutputfile(options, fulloutputpath):
        """Opens the output file."""
        if fulloutputpath is None:
            return StdoutWrapper()
        return open(fulloutputpath, "wb")

    @staticmethod
    def opentempoutputfile(options, fulloutputpath):
        """Opens a temporary output file."""
        return BytesIO()

    def finalizetempoutputfile(self, options, outputfile, fulloutputpath) -> None:
        """Write the temp outputfile to its final destination."""
        outputfile.seek(0, 0)
        outputstring = outputfile.read()
        outputfile = self.openoutputfile(options, fulloutputpath)
        outputfile.write(outputstring)
        outputfile.close()

    def opentemplatefile(self, options, fulltemplatepath):
        """Opens the template file (if required)."""
        if fulltemplatepath is not None:
            if os.path.isfile(fulltemplatepath):
                return open(fulltemplatepath, "rb")
            self.warning(f"missing template file {fulltemplatepath}")
        return None

    def processfile(
        self, fileprocessor, options, fullinputpath, fulloutputpath, fulltemplatepath
    ) -> bool:
        """Process an individual file."""
        inputfile = self.openinputfile(options, fullinputpath)
        if fulloutputpath and fulloutputpath in {fullinputpath, fulltemplatepath}:
            outputfile = self.opentempoutputfile(options, fulloutputpath)
            tempoutput = True
        else:
            outputfile = self.openoutputfile(options, fulloutputpath)
            tempoutput = False
        templatefile = self.opentemplatefile(options, fulltemplatepath)
        passthroughoptions = self.getpassthroughoptions(options)
        result = fileprocessor(
            inputfile, outputfile, templatefile, **passthroughoptions
        )
        if fullinputpath is not None:
            inputfile.close()
        if result:
            if tempoutput:
                self.warning("writing to temporary output...")
                self.finalizetempoutputfile(options, outputfile, fulloutputpath)
            if fulloutputpath and os.path.isfile(fulloutputpath):
                outputfile.close()
            return True
        # remove the file if it is a file (could be stdout etc)
        if fulloutputpath and os.path.isfile(fulloutputpath):
            outputfile.close()
            os.unlink(fulloutputpath)
        return False

    @staticmethod
    def mkdir(parent, subdir) -> None:
        """Makes a subdirectory (recursively if necessary)."""
        if not os.path.isdir(parent):
            raise ValueError(
                f"cannot make child directory {subdir!r} if parent {parent!r} does not exist"
            )
        subdir = RecursiveOptionParser.normalizerelativepath(subdir, allow_empty=True)
        if not subdir:
            return
        currentpath = parent
        subparts = subdir.split(os.sep)
        for part in subparts:
            currentpath = os.path.join(currentpath, part)
            if not os.path.isdir(currentpath):
                os.mkdir(currentpath)

    def checkoutputsubdir(self, options, subdir) -> None:
        """
        Checks to see if subdir under options.output needs to be created,
        creates if necessary.
        """
        subdir = self.normalizerelativepath(subdir, allow_empty=True)
        if not subdir:
            return
        fullpath = os.path.join(options.output, subdir)
        if not os.path.isdir(fullpath):
            self.mkdir(options.output, subdir)

    @staticmethod
    def isexcluded(options, inputpath) -> bool:
        """Checks if this path has been excluded."""
        basename = os.path.basename(inputpath)
        for excludename in options.exclude:
            if fnmatch.fnmatch(basename, excludename):
                return True
        return False

    def recurseinputfilelist(self, options):
        """Use a list of files, and find a common base directory for them."""
        # find a common base directory for the files to do everything
        # relative to
        commondir = os.path.dirname(os.path.commonpath(options.input))
        inputfiles = []
        for inputfile in options.input:
            if self.isexcluded(options, inputfile):
                continue
            if inputfile.startswith(commondir + os.sep):
                inputfiles.append(inputfile.replace(commondir + os.sep, "", 1))
            else:
                inputfiles.append(inputfile.replace(commondir, "", 1))
        options.input = commondir
        return inputfiles

    def recurseinputfiles(self, options):
        """Recurse through directories and return files to be processed."""
        dirstack = [""]
        join = os.path.join
        inputfiles = []
        while dirstack:
            top = dirstack.pop(-1)
            names = os.listdir(join(options.input, top))
            dirs = []
            for name in names:
                inputpath = join(top, name)
                if self.isexcluded(options, inputpath):
                    continue
                fullinputpath = self.getfullinputpath(options, inputpath)
                # handle directories...
                if os.path.isdir(fullinputpath):
                    dirs.append(inputpath)
                elif os.path.isfile(fullinputpath):
                    if not self.isvalidinputname(name):
                        # only handle names that match recognized input
                        # file extensions
                        continue
                    inputfiles.append(inputpath)
            # make sure the directories are processed next time round.
            dirs.reverse()
            dirstack.extend(dirs)
        return inputfiles

    @staticmethod
    def splitext(pathname: str) -> tuple[str, str]:
        """
        Splits *pathname* into name and ext, and removes the extsep.

        :param pathname: A file path
        :return: root, ext
        """
        root, ext = os.path.splitext(pathname)
        ext = ext.replace(os.extsep, "", 1)
        return (root, ext)

    def splitinputext(self, inputpath):
        """Splits an *inputpath* into name and extension."""
        return self.splitext(inputpath)

    def splittemplateext(self, templatepath):
        """Splits a *templatepath* into name and extension."""
        return self.splitext(templatepath)

    def templateexists(self, options, templatepath):
        """Returns whether the given template exists..."""
        fulltemplatepath = self.getfulltemplatepath(options, templatepath)
        return os.path.isfile(fulltemplatepath)

    def gettemplatename(self, options, inputname):
        """Gets an output filename based on the input filename."""
        if not self.usetemplates:
            return None
        if not inputname or not options.recursivetemplate:
            return options.template
        inputbase, inputext = self.splitinputext(inputname)
        if options.template:
            for inputext1, templateext1 in self.outputoptions:
                if inputext == inputext1 and templateext1:
                    templatepath = inputbase + os.extsep + templateext1
                    if self.templateexists(options, templatepath):
                        return templatepath
            if "*" in self.inputformats:
                for inputext1, templateext1 in self.outputoptions:
                    if inputext1 in {inputext, "*"}:
                        if templateext1 == "*":
                            templatepath = inputname
                            if self.templateexists(options, templatepath):
                                return templatepath
                        elif templateext1:
                            templatepath = inputbase + os.extsep + templateext1
                            if self.templateexists(options, templatepath):
                                return templatepath
        return None

    def getoutputname(self, options, inputname, outputformat):
        """Gets an output filename based on the input filename."""
        if not inputname or not options.recursiveoutput:
            return options.output
        inputbase, _inputext = self.splitinputext(inputname)
        outputname = inputbase
        if outputformat:
            outputname += os.extsep + outputformat
        return self.normalizerelativepath(outputname)

    def isvalidinputname(self, inputname):
        """Checks if this is a valid input filename."""
        _inputbase, inputext = self.splitinputext(inputname)
        return (inputext in self.inputformats) or ("*" in self.inputformats)



#
# Copyright 2004-2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
Module for parsing Qt .ts files for translation.

Currently this module supports the old format of .ts files. Some applications
use the newer .ts format which are documented here:
`TS file format 4.3 <http://doc.qt.io/archives/4.3/linguist-ts-file-format.html>`_,
`Example <http://svn.ez.no/svn/ezcomponents/trunk/Translation/docs/linguist-format.txt>`_

`Specification of the valid variable entries <http://doc.qt.io/qt-5/qstring.html#arg>`_,
`2 <http://doc.qt.io/qt-5/qstring.html#arg-2>`_
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, BinaryIO, TextIO

from lxml import etree

from translate.misc.xml_helpers import parse_xml
from translate.storage import base

if TYPE_CHECKING:
    from collections.abc import Iterator

TSInput = str | bytes | BinaryIO | TextIO
MessageNode = etree._Element
ContextNode = etree._Element
XML_DECLARATION_RE = re.compile(
    r"""^\s*<\?xml(?P<attrs>[^?]*?)\?>""",
    re.IGNORECASE | re.DOTALL,
)


class QtTsParser:
    def __init__(self, inputfile: TSInput | None = None) -> None:
        """Make a new QtTsParser, reading from the given inputfile if required."""
        input_name = None
        if inputfile is not None and not isinstance(inputfile, str):
            input_name = base.get_input_name(inputfile, include_filename_attr=True)
        self.filename = input_name
        self.knowncontextnodes: dict[str, ContextNode] = {}
        self.doctype: str | None
        self.document: etree._ElementTree
        if inputfile is None:
            self.doctype = "<!DOCTYPE TS>"
            self.document = self._parse_content(b"<TS></TS>")
        else:
            prepared = base.prepare_input(inputfile)
            content = prepared.data
            if not prepared.from_handle and base.is_path_input(content):
                content = base.path_input_str(content)
                if not self._looks_like_xml(content):
                    self.filename = content
                    with open(content, "rb") as input_handle:
                        content = input_handle.read()
            self.document = self._parse_content(content)
            # Preserve the original doctype so round-tripped TS files keep the
            # same declaration shape as the input.
            self.doctype = self.document.docinfo.doctype or None
            assert self.document.getroot().tag == "TS"

    @staticmethod
    def _parse_content(content: str | bytes) -> etree._ElementTree:
        if isinstance(content, str):
            content = QtTsParser._normalize_text_input(content).encode("utf-8")
        root = parse_xml(content)
        return root.getroottree()

    @staticmethod
    def _looks_like_xml(content: str) -> bool:
        return content.lstrip().startswith("<")

    @staticmethod
    def _normalize_text_input(content: str) -> str:
        """
        Normalize text input before parsing as UTF-8 bytes.

        Text streams are already decoded, so any XML encoding declaration no
        longer reflects the in-memory string. Drop just the encoding attribute
        to preserve the remaining declaration fields without corrupting text.
        """
        declaration = XML_DECLARATION_RE.match(content)
        if declaration is None:
            return content
        attrs = QtTsParser._strip_xml_encoding_attribute(declaration.group("attrs"))
        return f"<?xml{attrs}?>{content[declaration.end() :]}"

    @staticmethod
    def _strip_xml_encoding_attribute(attrs: str) -> str:
        result: list[str] = []
        index = 0
        attr_count = len(attrs)

        while index < attr_count:
            whitespace_start = index
            while index < attr_count and attrs[index].isspace():
                index += 1

            if index >= attr_count:
                result.append(attrs[whitespace_start:index])
                break

            name_start = index
            while index < attr_count and (
                attrs[index].isalnum() or attrs[index] in "._:-"
            ):
                index += 1
            if name_start == index:
                return attrs

            name = attrs[name_start:index]

            while index < attr_count and attrs[index].isspace():
                index += 1
            if index >= attr_count or attrs[index] != "=":
                return attrs
            index += 1

            while index < attr_count and attrs[index].isspace():
                index += 1
            if index >= attr_count or attrs[index] not in {'"', "'"}:
                return attrs

            quote = attrs[index]
            index += 1
            value_end = attrs.find(quote, index)
            if value_end == -1:
                return attrs
            index = value_end + 1

            if name.lower() != "encoding":
                result.append(attrs[whitespace_start:index])

        return "".join(result)

    @property
    def documentElement(self) -> etree._Element:
        return self.document.getroot()

    def addtranslation(
        self,
        contextname: str,
        source: str,
        translation: str,
        comment: str | None = None,
        transtype: str | None = None,
        createifmissing: bool = False,
    ) -> bool:
        """Adds the given translation (will create the nodes required if asked). Returns success."""
        contextnode = self.getcontextnode(contextname)
        if contextnode is None:
            if not createifmissing:
                return False
            contextnode = etree.SubElement(self.documentElement, "context")
            etree.SubElement(contextnode, "name").text = contextname
            self.knowncontextnodes[contextname] = contextnode
        if not createifmissing:
            return False
        messagenode = etree.SubElement(contextnode, "message")
        etree.SubElement(messagenode, "source").text = source
        if comment:
            etree.SubElement(messagenode, "comment").text = comment
        translationnode = etree.SubElement(messagenode, "translation")
        translationnode.text = translation
        if transtype:
            translationnode.set("type", transtype)
        return True

    def getxml(self) -> str:
        """Return the ts file as xml."""
        xml = etree.tostring(
            self.document,
            pretty_print=True,
            encoding="utf-8",
            xml_declaration=True,
            doctype=self.doctype,
        ).decode("utf-8")
        # Keep output compact and match the historic behavior of dropping
        # blank lines introduced by pretty printing.
        return "\n".join(line for line in xml.split("\n") if line.strip())

    @staticmethod
    def getcontextname(contextnode: ContextNode) -> str:
        """Returns the name of the given context."""
        return contextnode.findtext("name", default="")

    def getcontextnode(self, contextname: str) -> ContextNode | None:
        """Finds the contextnode with the given name."""
        contextnode = self.knowncontextnodes.get(contextname)
        if contextnode is not None:
            return contextnode
        # Cache by context name because repeated lookups are common during
        # addtranslation() and iterative reads.
        for contextnode in self.documentElement.findall("context"):
            if self.getcontextname(contextnode) == contextname:
                self.knowncontextnodes[contextname] = contextnode
                return contextnode
        return None

    def getmessagenodes(
        self, context: str | ContextNode | None = None
    ) -> list[MessageNode]:
        """Returns all the messagenodes, limiting to the given context (name or node) if given."""
        if context is None:
            return self.documentElement.findall("./context/message")
        if isinstance(context, str):
            context = self.getcontextnode(context)
            if context is None:
                return []
        return context.findall("message")

    @staticmethod
    def getmessagesource(message: MessageNode) -> str:
        """Returns the message source for a given node."""
        return message.findtext("source", default="")

    @staticmethod
    def getmessagetranslation(message: MessageNode) -> str:
        """Returns the message translation for a given node."""
        return message.findtext("translation", default="")

    @staticmethod
    def getmessagetype(message: MessageNode) -> str:
        """Returns the message translation attributes for a given node."""
        translationnode = message.find("translation")
        if translationnode is None:
            return ""
        return translationnode.get("type", "")

    @staticmethod
    def getmessagecomment(message: MessageNode) -> str:
        """Returns the message comment for a given node."""
        return message.findtext("comment", default="")

    def iteritems(self) -> Iterator[tuple[str, list[MessageNode]]]:
        """Iterates through (contextname, messages)."""
        for contextnode in self.documentElement.findall("context"):
            yield self.getcontextname(contextnode), self.getmessagenodes(contextnode)



#
# Copyright 2023 Zuza Software Foundation & Anders Kaplan
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Translate Markdown files using Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/md2po.html
for examples and usage instructions.
"""

from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING

from translate.convert import convert
from translate.misc.optrecurse import ProgressBar
from translate.storage import markdown, po

if TYPE_CHECKING:
    from translate.storage.base import TranslationStore

DEFAULT_MAX_LINE_LENGTH = 80


class MarkdownTranslator:
    def __init__(
        self,
        inputstore: TranslationStore,
        includefuzzy: bool,
        outputthreshold: int | None,
        maxlength: int,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ) -> None:
        self.inputstore = inputstore
        self.inputstore.require_index()
        self.includefuzzy = includefuzzy
        self.outputthreshold = outputthreshold
        self.maxlength = maxlength
        self.extract_code_blocks = extract_code_blocks
        self.extract_frontmatter = extract_frontmatter
        self.no_placeholders = no_placeholders

    def translate(self, templatefile, outputfile) -> int:
        if not convert.should_output_store(self.inputstore, self.outputthreshold):
            return False

        outputstore = markdown.MarkdownFile(
            inputfile=templatefile,
            callback=self._lookup,
            max_line_length=self.maxlength if self.maxlength > 0 else None,
            extract_code_blocks=self.extract_code_blocks,
            extract_frontmatter=self.extract_frontmatter,
            no_placeholders=self.no_placeholders,
        )
        outputfile.write(outputstore.filesrc.encode("utf-8"))
        return 1

    def _lookup(self, string: str) -> str:
        unit = self.inputstore.sourceindex.get(string, None)
        if unit is None:
            return string
        unit = unit[0]
        if unit.istranslated():
            return unit.target
        if self.includefuzzy and unit.isfuzzy():
            return unit.target
        return unit.source


class PO2MDOptionParser(convert.ConvertOptionParser):
    def __init__(self) -> None:
        formats = {
            ("po", "md"): ("md", self._translate_md_file),
            ("po", "markdown"): ("markdown", self._translate_md_file),
            ("po", "txt"): ("txt", self._translate_md_file),
            ("po", "text"): ("text", self._translate_md_file),
        }
        super().__init__(formats, usetemplates=True, description=__doc__)
        self.add_option(
            "-m",
            "--maxlinelength",
            type="int",
            dest="maxlength",
            default=DEFAULT_MAX_LINE_LENGTH,
            help="reflow (word wrap) the output to the given maximum line length. set to 0 to disable",
        )
        self.passthrough.append("maxlength")
        self.add_option(
            "",
            "--no-code-blocks",
            action="store_false",
            dest="extract_code_blocks",
            default=True,
            help="do not extract code blocks for translation",
        )
        self.passthrough.append("extract_code_blocks")
        self.add_option(
            "",
            "--no-frontmatter",
            action="store_false",
            dest="extract_frontmatter",
            default=True,
            help="do not extract front matter for translation",
        )
        self.passthrough.append("extract_frontmatter")
        self.add_option(
            "",
            "--no-placeholders",
            action="store_true",
            dest="no_placeholders",
            default=False,
            help=(
                "look up translation units by their full inline markdown "
                "(links, images, autolinks, HTML spans rendered verbatim) "
                "instead of {n} placeholder markers; use together with "
                "md2po --no-placeholders"
            ),
        )
        self.passthrough.append("no_placeholders")
        self.add_threshold_option()
        self.add_fuzzy_option()

    @staticmethod
    def _translate_md_file(
        inputfile,
        outputfile,
        templatefile,
        includefuzzy: bool,
        outputthreshold: int | None,
        maxlength: int,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ):
        inputstore = po.pofile(inputfile)
        translator = MarkdownTranslator(
            inputstore,
            includefuzzy,
            outputthreshold,
            maxlength,
            extract_code_blocks=extract_code_blocks,
            extract_frontmatter=extract_frontmatter,
            no_placeholders=no_placeholders,
        )
        return translator.translate(templatefile, outputfile)

    def recursiveprocess(self, options) -> None:
        if (
            self.isrecursive(options.template, "template")
            and not self.isrecursive(options.input, "input")
            and self.can_be_recursive(options.output, "output")
        ):
            self.recursiveprocess_by_templates(options)
        else:
            super().recursiveprocess(options)

    def can_be_recursive(self, fileoption, filepurpose):
        return fileoption is not None and not os.path.isfile(fileoption)

    def recursiveprocess_by_templates(self, options) -> None:
        """Recurse through directories and process files, by templates (markdown) not input files (po)."""
        inputfile = self.openinputfile(options, options.input)
        self.inputstore = po.pofile(inputfile)
        templatefiles = self.recurse_template_files(options)
        self.ensurerecursiveoutputdirexists(options)
        progress_bar = ProgressBar(options.progress, templatefiles)
        for templatepath in templatefiles:
            fulltemplatepath = os.path.join(options.template, templatepath)
            outputpath = templatepath
            fulloutputpath = os.path.join(options.output, outputpath)
            self.checkoutputsubdir(options, os.path.dirname(outputpath))
            try:
                success = self.processfile(
                    self.process_file_with_fixed_inputstore,
                    options,
                    None,
                    fulloutputpath,
                    fulltemplatepath,
                )
            except Exception:
                self.warning(
                    f"Error processing: input {options.input}, output {fulloutputpath}, template {fulltemplatepath}",
                    options,
                    sys.exc_info(),
                )
                success = False
            progress_bar.report_progress(templatepath, success)
        del progress_bar

    def process_file_with_fixed_inputstore(
        self,
        inputfile,
        outputfile,
        templatefile,
        includefuzzy: bool,
        outputthreshold: int | None,
        maxlength: int,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ):
        translator = MarkdownTranslator(
            self.inputstore,
            includefuzzy,
            outputthreshold,
            maxlength,
            extract_code_blocks=extract_code_blocks,
            extract_frontmatter=extract_frontmatter,
            no_placeholders=no_placeholders,
        )
        return translator.translate(templatefile, outputfile)

    def recurse_template_files(self, options):
        """Recurse through directories and return files to be processed."""
        dirstack = [""]
        join = os.path.join
        templatefiles = []
        while dirstack:
            top = dirstack.pop(-1)
            names = os.listdir(join(options.template, top))
            dirs = []
            for name in names:
                filepath = join(top, name)
                fullfilepath = join(options.template, filepath)
                # handle directories...
                if os.path.isdir(fullfilepath):
                    dirs.append(filepath)
                elif os.path.isfile(fullfilepath):
                    if not self.isvalidtemplatename(name):
                        # only handle names that match recognized output
                        # file extensions
                        continue
                    templatefiles.append(filepath)
            # make sure the directories are processed next time round.
            dirs.reverse()
            dirstack.extend(dirs)
        return templatefiles

    _txt_extensions = {"txt", "text"}

    def isvalidtemplatename(self, filename):
        """Checks if this is a valid template/output filename."""
        _, ext = self.splitext(filename)
        if ext in self._txt_extensions:
            return False
        return any(ext == templateformat for _, templateformat in self.outputoptions)


def main(argv=None) -> None:
    parser = PO2MDOptionParser()
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2003-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert Gettext PO localization files to Comma-Separated Value (.csv) files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/csv2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.storage import csvl10n, po


class po2csv:
    @staticmethod
    def convertcomments(inputunit):
        return " ".join(inputunit.getlocations())

    def convertunit(self, inputunit):
        if inputunit.isheader():
            return None
        if inputunit.isblank():
            return None
        csvunit = csvl10n.csvunit()
        csvunit.location = self.convertcomments(inputunit)
        csvunit.source = inputunit.source
        csvunit.target = inputunit.target
        csvunit.setcontext(inputunit.getcontext())
        return csvunit

    def convertplurals(self, inputunit):
        """
        Convert PO plural units.

        We only convert the first plural form.  So languages with multiple
        plurals are not handled.  For single plural languages we simply
        skip this plural extraction.
        """
        if len(inputunit.target.strings) == 1:  # No plural forms
            return None
        csvunit = csvl10n.csvunit()
        csvunit.location = self.convertcomments(inputunit)
        csvunit.source = inputunit.source.strings[1]
        csvunit.target = inputunit.target.strings[1]
        return csvunit

    def convertstore(self, inputstore, columnorder=None):
        if columnorder is None:
            columnorder = ["location", "source", "target"]
        outputstore = csvl10n.csvfile(fieldnames=columnorder)
        for inputunit in inputstore.units:
            outputunit = self.convertunit(inputunit)
            if outputunit is not None:
                outputstore.addunit(outputunit)
            if inputunit.hasplural():
                outputunit = self.convertplurals(inputunit)
                if outputunit is not None:
                    outputstore.addunit(outputunit)
        return outputstore


def convertcsv(inputfile, outputfile, templatefile, columnorder=None) -> int:
    """Reads in inputfile using po, converts using po2csv, writes to outputfile."""
    # note that templatefile is not used, but it is required by the converter...
    inputstore = po.pofile(inputfile)
    if inputstore.isempty():
        return 0
    convertor = po2csv()
    outputstore = convertor.convertstore(inputstore, columnorder)
    outputstore.serialize(outputfile)
    return 1


def columnorder_callback(option, opt, value, parser) -> None:
    setattr(parser.values, option.dest, value.split(","))


def main(argv=None) -> None:
    formats = {"po": ("csv", convertcsv)}
    parser = convert.ConvertOptionParser(formats, description=__doc__)
    parser.add_option(
        "",
        "--columnorder",
        dest="columnorder",
        action="callback",
        callback=columnorder_callback,
        type="str",
        default=None,
        help="specify the order and position of columns (location,source,target,context)",
    )
    parser.passthrough.append("columnorder")
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2012 Michal Čihař
# Copyright 2014 Luca De Petrillo
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Module for handling Android String and Plurals resource files."""

from __future__ import annotations

import copy
import os
import re
import warnings
from typing import ClassVar, cast, overload
from xml.parsers.expat import XML_PARAM_ENTITY_PARSING_NEVER, ParserCreate

from lxml import etree

from translate.lang import data
from translate.misc.multistring import multistring
from translate.misc.xml_helpers import get_safe_xml_parser
from translate.storage import base, lisa

WHITESPACE = {" ", "\n", "\t"}  # Whitespace that we collapse.
ESCAPE_NEWLINE = {"n", "N"}
ESCAPE_TAB = {"t", "T"}
ESCAPE_PLAIN = {" ", '"', "'", "@", "?", "\\"}
MULTIWHITESPACE = re.compile(r"[ \n\t]{2}(?!\\n)")
QUOTED_STRING = re.compile(r'((?<!\\)"(?:\\"|[^"])*(?<!\\)")')
UNICODE_ESCAPE = re.compile(r"\\u([a-fA-F0-9]{4})")
CHAR_ESCAPE = re.compile(r"\\(.)")
WHITESPACE_RE = re.compile(r"\s+")
XML_NAME = r"[A-Za-z_][A-Za-z0-9_.:-]*"
XML_ATTRIBUTE = rf'\s+{XML_NAME}\s*=\s*(?:"[^"]*"|\'[^\']*\')'
XML_ATTRIBUTE_CAPTURE = re.compile(rf"({XML_NAME})\s*=\s*(\"[^\"]*\"|'[^']*')")
XML_TAG_NAME = re.compile(rf"^</?({XML_NAME})")
MARKUP_TAG = re.compile(rf"</?{XML_NAME}(?:{XML_ATTRIBUTE})*\s*/?>")
ESCAPED_MARKUP_TAG = re.compile(rf"&lt;/?{XML_NAME}(?:{XML_ATTRIBUTE})*\s*/?&gt;")
CDATA_ONLY = re.compile(r"(?:<!\[CDATA\[.*?\]\]>)+", re.DOTALL)


class DecodingXMLParser:
    """
    XML parser that processes non CDATA strings.

    - all XML markup from given depth is output as is
    - CDATA elements are output as is
    - process_string() is called on all other textual content
    """

    EMIT_DEPTH = 1
    FOREIGN_DTD = True

    def __init__(self, text: str, *, escape_all: bool = True) -> None:
        self.text = text.encode("utf-8")
        self.output: list[str] = []
        self.emit_start: int | None = None
        self.character_data = False
        self.raw_string = True
        self.in_string = False
        self.inside_quoted = False
        self.do_cleanup = False
        self.depth = 0
        self.escape_all = escape_all
        self.parser = parser = ParserCreate()
        if self.FOREIGN_DTD:
            parser.UseForeignDTD(self.FOREIGN_DTD)
        parser.SetParamEntityParsing(XML_PARAM_ENTITY_PARSING_NEVER)
        parser.StartElementHandler = self.StartElementHandler
        parser.EndElementHandler = self.EndElementHandler
        parser.CharacterDataHandler = self.CharacterDataHandler
        parser.StartCdataSectionHandler = self.StartCdataSectionHandler
        parser.EndCdataSectionHandler = self.EndCdataSectionHandler
        parser.DefaultHandler = self.DefaultHandler

    @staticmethod
    def decode_unicode_escapes(match: re.Match) -> str:
        escaped = match.group(1)
        # Convert from hex to character
        return chr(int(escaped, 16))

    def decode_escapes(self, match: re.Match) -> str:
        escaped = match.group(1)
        if escaped in ESCAPE_NEWLINE:
            return "\n"  # an actual newline
        if escaped in ESCAPE_TAB:
            return "\t"  # an actual tab
        if escaped == "\\":
            return "\\"
        # Compose Multiplatform Resources handles escaping only for few chars
        if not self.escape_all:
            return rf"\{escaped}"
        if escaped in ESCAPE_PLAIN:
            # Plain string to escape
            return escaped
        # All others, remove, like Android does as well.
        return ""

    def strip_part(self, text: str) -> tuple[str, bool]:
        # Detect leading quote
        if text.startswith('"'):
            text = text[1:]
            self.inside_quoted ^= True

        # Check if whitespace should be stripped
        strip_whitespace = not self.inside_quoted

        # Detect trailing quote
        if text.endswith('"') and not text.endswith('\\"'):
            text = text[:-1]
            self.inside_quoted ^= True

        # Consolidate whitespace if needed
        if strip_whitespace:
            text = WHITESPACE_RE.sub(" ", text)

        return text, strip_whitespace

    def process_string(self, content: str) -> tuple[str, bool, bool]:
        # Skip processing for a blank string
        if not content:
            return "", False, False

        # Extract quoted parts, remove blank ones
        parts = [part for part in QUOTED_STRING.split(content) if part]

        # Detect possible cleanups
        cleanup_start = False
        cleanup_end = False

        # Collapse whitespace
        stripped_parts: list[str] = []
        for i, part in enumerate(parts):
            stripped, was_stripped = self.strip_part(part)
            if i == 0:
                cleanup_start = was_stripped
            cleanup_end = was_stripped and not self.inside_quoted
            stripped_parts.append(stripped)

        # Merge this back to a single string
        text = "".join(stripped_parts)

        # Handle unicode escapes
        unescaped_text = UNICODE_ESCAPE.sub(self.decode_unicode_escapes, text)
        # Handle other escapes
        unescaped_text = CHAR_ESCAPE.sub(self.decode_escapes, unescaped_text)

        # Detect escaping at the start and end of a string (we actually care only for
        # whitespace escapes, but detecting more does not matter)
        if text and unescaped_text and text[0] != unescaped_text[0]:
            cleanup_start = False
        if text and unescaped_text and text[-1] != unescaped_text[-1]:
            cleanup_end = False

        return unescaped_text, cleanup_start, cleanup_end

    def cleanup_text(self, text: str) -> str:
        """Remove leading whitespace from text."""
        if not self.output:
            return text.lstrip()
        return text

    def emit(self) -> None:
        if self.emit_start is not None:
            text = self.text[self.emit_start : self.parser.CurrentByteIndex].decode(
                "utf-8"
            )
            self.do_cleanup = not self.raw_string
            if not self.raw_string:
                text, cleanup_start, self.do_cleanup = self.process_string(text)
                if cleanup_start:
                    text = self.cleanup_text(text)
            self.output.append(text)
            self.emit_start = None
        self.character_data = False
        self.raw_string = True
        self.in_string = False

    def StartElementHandler(self, _name, _attributes) -> None:
        self.emit()
        if self.depth >= self.EMIT_DEPTH:
            self.emit_start = self.parser.CurrentByteIndex
        self.depth += 1

    def EndElementHandler(self, _name) -> None:
        self.emit()
        if self.depth >= self.EMIT_DEPTH:
            self.emit_start = self.parser.CurrentByteIndex
        self.depth -= 1

    def CharacterDataHandler(self, _data) -> None:
        if not self.character_data and not self.in_string:
            self.emit()
            self.emit_start = self.parser.CurrentByteIndex
            self.raw_string = False
            self.in_string = True

    def StartCdataSectionHandler(self) -> None:
        self.emit()
        self.character_data = True
        self.emit_start = self.parser.CurrentByteIndex

    def EndCdataSectionHandler(self) -> None:
        self.character_data = False

    def DefaultHandler(self, _data) -> None:
        if self.depth >= self.EMIT_DEPTH:
            self.emit()
            self.emit_start = self.parser.CurrentByteIndex

    def parse(self) -> str:
        self.parser.Parse(self.text, True)
        if self.depth >= self.EMIT_DEPTH:
            self.emit()
        # Remove trailing whitespace from text
        if self.do_cleanup and self.output:
            self.output[-1] = self.output[-1].rstrip()
        return "".join(self.output)


class EncodingXMLParser(DecodingXMLParser):
    EMIT_DEPTH = 0
    FOREIGN_DTD = False

    def process_string(self, text: str) -> tuple[str, bool, bool]:  # ty:ignore[invalid-method-override]
        return (
            AndroidResourceUnit.escape(text, quote_wrapping_whitespaces=False),
            False,
            False,
        )

    def parse(self) -> str:
        self.emit_start = 0
        return super().parse()


class AndroidResourceUnit(base.TranslationUnit):
    """A single entry in the Android String resource file."""

    SINGULAR_TAG = "string"
    PLURAL_TAG = "plurals"
    ESCAPE_TRANSLATE: ClassVar = str.maketrans(
        {
            "\\": "\\\\",
            "\n": "\\n",
            "\t": "\\t",
            "'": "\\'",
            '"': '\\"',
        }
    )
    ESCAPE_ALL: ClassVar[bool] = True
    TARGET_MARKUP_PLAIN: ClassVar[str] = "plain"
    TARGET_MARKUP_ESCAPED: ClassVar[str] = "escaped-markup"
    TARGET_MARKUP_CDATA: ClassVar[str] = "cdata"
    TARGET_MARKUP_XML: ClassVar[str] = "xml-markup"

    TargetType = str | multistring | list[str] | None

    @classmethod
    def createfromxmlElement(cls, element: etree._Element):
        term = None
        # Actually this class supports only plurals and string tags
        if element.tag in {cls.PLURAL_TAG, cls.SINGULAR_TAG}:
            term = cls(None, xmlelement=element)
        return term

    def __init__(
        self,
        source,
        empty: bool = False,
        xmlelement: etree._Element | None = None,
        **kwargs,
    ) -> None:
        if xmlelement is not None:
            self.xmlelement = xmlelement
            self._target_markup_mode = self.TARGET_MARKUP_PLAIN
            self._init_target_markup_mode()
        elif self.hasplurals(source):
            self.xmlelement = etree.Element(self.PLURAL_TAG)
            self._target_markup_mode = self.TARGET_MARKUP_XML
        else:
            self.xmlelement = etree.Element(self.SINGULAR_TAG)
            self._target_markup_mode = self.TARGET_MARKUP_XML
        if source is not None:
            self.setid(source)
        super().__init__(source)

    @staticmethod
    def _contains_markup_text(text: str | None) -> bool:
        return bool(text) and bool(MARKUP_TAG.search(text))

    @staticmethod
    def _serialize_text(text: str) -> str:
        wrapper = etree.Element("wrapper")
        wrapper.text = text
        serialized = etree.tostring(wrapper, encoding="unicode", with_tail=False)
        return serialized.removeprefix("<wrapper>").removesuffix("</wrapper>")

    @staticmethod
    def _unescape_markup_tag(match: re.Match) -> str:
        return match.group(0).replace("&lt;", "<").replace("&gt;", ">")

    def _render_xml_context(self, xmltarget: etree._Element, target: str) -> str:
        probe_marker = "__translate_toolkit_markup_probe__"
        probe_target = copy.deepcopy(xmltarget)
        for child in list(probe_target):
            probe_target.remove(child)
        probe_target.text = probe_marker
        probe_target.tail = None

        source_document = self._store.document if self._store is not None else None
        if source_document is None:
            source_document = xmltarget.getroottree()

        source_root = source_document.getroot()
        if source_root is xmltarget:
            template = etree.tostring(probe_target, encoding="unicode", with_tail=False)
            return template.replace(probe_marker, target)

        cloned_doc = copy.deepcopy(source_document)
        cloned_root = cloned_doc.getroot()
        cloned_root.clear()
        cloned_root.text = " "
        cloned_root.append(probe_target)
        template = etree.tostring(
            cloned_doc,
            encoding="unicode",
            doctype=copy.copy(self._store.XMLdoctype)
            if self._store is not None
            else None,
        )
        return template.replace(probe_marker, target)

    def _get_markup_scope(self, xmltarget: etree._Element) -> dict[str | None, str]:
        source_document = self._store.document if self._store is not None else None
        if source_document is None:
            source_document = xmltarget.getroottree()
        source_root = source_document.getroot()
        scope = {
            prefix: namespace
            for prefix, namespace in source_root.nsmap.items()
            if namespace is not None
        }
        scope.update(
            {
                prefix: namespace
                for prefix, namespace in xmltarget.nsmap.items()
                if namespace is not None
            }
        )
        return scope

    @staticmethod
    def _get_markup_name(tag: str) -> str | None:
        match = XML_TAG_NAME.match(tag)
        if match is None:
            return None
        return match.group(1)

    @staticmethod
    def _update_markup_scope(
        scope: dict[str | None, str],
        tag: str,
    ) -> dict[str | None, str]:
        tag_scope = scope.copy()
        for attribute, value in XML_ATTRIBUTE_CAPTURE.findall(tag):
            namespace = value[1:-1]
            if attribute == "xmlns":
                tag_scope[None] = namespace
            elif attribute.startswith("xmlns:"):
                tag_scope[attribute.split(":", 1)[1]] = namespace
        return tag_scope

    @staticmethod
    def _is_resolved_markup_name(
        markup_name: str,
        scope: dict[str | None, str],
    ) -> bool:
        if ":" not in markup_name:
            return True
        prefix = markup_name.split(":", 1)[0]
        return prefix in scope

    def _has_escaped_markup_structure(
        self, xmltarget: etree._Element, text: str
    ) -> bool:
        scope_stack = [self._get_markup_scope(xmltarget)]
        open_tags: list[str] = []
        saw_tag = False

        for match in ESCAPED_MARKUP_TAG.finditer(text):
            tag = self._unescape_markup_tag(match)
            markup_name = self._get_markup_name(tag)
            if markup_name is None:
                return False

            if tag.startswith("</"):
                if not open_tags or open_tags[-1] != markup_name:
                    return False
                if not self._is_resolved_markup_name(markup_name, scope_stack[-1]):
                    return False
                open_tags.pop()
                scope_stack.pop()
                saw_tag = True
                continue

            tag_scope = self._update_markup_scope(scope_stack[-1], tag)
            if not self._is_resolved_markup_name(markup_name, tag_scope):
                return False

            saw_tag = True
            if tag.endswith("/>"):
                continue

            open_tags.append(markup_name)
            scope_stack.append(tag_scope)

        return saw_tag and not open_tags

    def _uses_plain_markup_roundtrip(self, xmltarget: etree._Element) -> bool:
        raw_xml = etree.tostring(xmltarget, encoding="unicode", with_tail=False)
        if len(xmltarget) != 0 or "<![CDATA[" in raw_xml:
            return False
        text = xmltarget.text
        if text is None or not self._contains_markup_text(text):
            return False
        serialized_text = self._serialize_text(text)
        parsed_text = ESCAPED_MARKUP_TAG.sub(self._unescape_markup_tag, serialized_text)
        if parsed_text == serialized_text:
            return False
        return self._has_escaped_markup_structure(xmltarget, serialized_text)

    @staticmethod
    def _has_cdata_text(xmltarget: etree._Element) -> bool:
        raw_xml = etree.tostring(xmltarget, encoding="unicode", with_tail=False)
        return len(xmltarget) == 0 and "<![CDATA[" in raw_xml

    @staticmethod
    def _extract_inner_xml(raw_xml: str) -> str:
        start = raw_xml.find(">")
        end = raw_xml.rfind("<")
        if start == -1 or end == -1 or end <= start:
            return ""
        return raw_xml[start + 1 : end]

    @classmethod
    def _has_only_cdata_sections(cls, raw_xml: str) -> bool:
        return bool(CDATA_ONLY.fullmatch(cls._extract_inner_xml(raw_xml)))

    @staticmethod
    def _has_literal_angle_text(xmltarget: etree._Element) -> bool:
        raw_xml = etree.tostring(xmltarget, encoding="unicode", with_tail=False)
        return (
            len(xmltarget) == 0
            and "<![CDATA[" not in raw_xml
            and xmltarget.text is not None
            and "<" in xmltarget.text
            and ">" in xmltarget.text
        )

    def _get_markup_targets(self) -> list[etree._Element]:
        if self.xmlelement.tag == self.PLURAL_TAG:
            return list(self.xmlelement.iterchildren("item"))
        return [self.xmlelement]

    def _has_xml_markup(self) -> bool:
        return any(len(xmltarget) != 0 for xmltarget in self._get_markup_targets())

    def _init_target_markup_mode(self) -> None:
        if self._has_xml_markup():
            self._target_markup_mode = self.TARGET_MARKUP_XML
        elif any(
            self._has_cdata_text(xmltarget) for xmltarget in self._get_markup_targets()
        ):
            self._target_markup_mode = self.TARGET_MARKUP_CDATA
        elif any(
            self._uses_plain_markup_roundtrip(xmltarget)
            for xmltarget in self._get_markup_targets()
        ):
            self._target_markup_mode = self.TARGET_MARKUP_ESCAPED
        elif any(
            self._has_literal_angle_text(xmltarget)
            for xmltarget in self._get_markup_targets()
        ):
            self._target_markup_mode = self.TARGET_MARKUP_PLAIN
        else:
            self._target_markup_mode = self.TARGET_MARKUP_XML

    def _get_target_markup_mode(self) -> str:
        return self._target_markup_mode

    def _set_target_markup_mode(self, target_markup_mode: str) -> None:
        if target_markup_mode not in {
            self.TARGET_MARKUP_PLAIN,
            self.TARGET_MARKUP_ESCAPED,
            self.TARGET_MARKUP_CDATA,
            self.TARGET_MARKUP_XML,
        }:
            msg = f"Unsupported target markup mode: {target_markup_mode}"
            raise ValueError(msg)
        self._target_markup_mode = target_markup_mode

    def gettargetmarkupmode(self) -> str:
        return self._get_target_markup_mode()

    def settargetmarkupmode(self, target_markup_mode: str) -> None:
        self._set_target_markup_mode(target_markup_mode)

    target_markup_mode = property(gettargetmarkupmode, settargetmarkupmode)

    def gettargetmarkup(self) -> bool:
        warnings.warn(
            "target_markup is deprecated; use target_markup_mode instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._get_target_markup_mode() == self.TARGET_MARKUP_ESCAPED

    def settargetmarkup(self, preserve_markup: bool) -> None:
        warnings.warn(
            "target_markup is deprecated; use target_markup_mode instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._set_target_markup_mode(
            self.TARGET_MARKUP_ESCAPED if preserve_markup else self.TARGET_MARKUP_XML
        )

    target_markup = property(gettargetmarkup, settargetmarkup)

    def istranslatable(self):
        return bool(self.getid()) and self.xmlelement.get("translatable") != "false"

    def marktranslatable(self, value: bool) -> None:
        return self.xmlelement.set("translatable", "true" if value else "false")

    def isblank(self) -> bool:
        return not bool(self.getid())

    def getid(self):
        return self.xmlelement.get("name")

    def setid(self, newid):  # ty:ignore[invalid-method-override]
        return self.xmlelement.set("name", newid)

    def getcontext(self):
        return self.xmlelement.get("name")

    @staticmethod
    def xml_escape_space(matchobj):
        return matchobj.group(0).replace("  ", r" \u0020")

    @overload
    @classmethod
    def escape(cls, text: str, quote_wrapping_whitespaces: bool = True) -> str: ...
    @overload
    @classmethod
    def escape(cls, text: None, quote_wrapping_whitespaces: bool = True) -> None: ...
    @classmethod
    def escape(cls, text, quote_wrapping_whitespaces=True):
        """
        Escape all the characters which need to be escaped in an Android XML
        file.

        :param text: Text to escape
        :param quote_wrapping_whitespaces: If True, heading and trailing
               whitespaces will be quoted placing the entire resulting text in
               double quotes.
        """
        if text is None:
            return None
        if len(text) == 0:
            return ""

        # Escape XML chars and whitespace
        text = text.translate(cls.ESCAPE_TRANSLATE)

        # @ and ? needs to be escaped at start as this would be interpreted
        # as string/style references
        if text.startswith(("@", "?")):
            text = f"\\{text}"

        # Quote strings with more whitespace
        multispace = MULTIWHITESPACE.search(text)
        if quote_wrapping_whitespaces and (
            text[0] in WHITESPACE or text[-1] in WHITESPACE or multispace
        ):
            return f'"{text}"'
        # In xml multispace
        if not quote_wrapping_whitespaces and multispace:
            return MULTIWHITESPACE.sub(cls.xml_escape_space, text)

        return text

    @base.TranslationUnit.source.getter
    def source(self):
        if super().source is None:
            return self.target
        return super().source

    def get_xml_text_value(self, xmltarget: etree._Element) -> str:
        raw_xml = etree.tostring(xmltarget, encoding="unicode", with_tail=False)
        # Use decoding parser to keep XML entities, CDATA while processing Android escaping
        parser = DecodingXMLParser(raw_xml, escape_all=self.ESCAPE_ALL)

        # Unescape as plain text in case there is no XML markup. Ufortunately
        # CDATA elements are not listed as children in lxml, so test for it in raw XML.
        if len(xmltarget) == 0:
            if "<![CDATA[" not in raw_xml:
                if xmltarget.text is None:
                    return ""
                text, cleanup_start, cleanup_end = parser.process_string(xmltarget.text)
                if cleanup_start:
                    text = text.lstrip()
                if cleanup_end:
                    text = text.rstrip()
                return text
            if self._has_only_cdata_sections(raw_xml):
                text = xmltarget.text or ""
                if self._contains_markup_text(text):
                    unescaped = UNICODE_ESCAPE.sub(parser.decode_unicode_escapes, text)
                    return CHAR_ESCAPE.sub(parser.decode_escapes, unescaped)
                text, cleanup_start, cleanup_end = parser.process_string(text)
                if cleanup_start:
                    text = text.lstrip()
                if cleanup_end:
                    text = text.rstrip()
                return text

        return parser.parse()

    def set_xml_text_plain(
        self,
        target: str | None,
        xmltarget: etree._Element,
        *,
        target_markup_mode: str = TARGET_MARKUP_PLAIN,
    ) -> None:
        # Remove possible old elements
        for x in xmltarget.iterchildren():
            xmltarget.remove(x)
        # Handle text only
        xmltarget.text = self.escape(target)
        self._set_target_markup_mode(target_markup_mode)

    def _parse_target_xml(
        self,
        xmltarget: etree._Element,
        target: str,
        *,
        escape_text_parts: bool = True,
    ) -> etree._Element | None:
        parser = get_safe_xml_parser(strip_cdata=False)
        newstring = self._render_xml_context(xmltarget, target)
        try:
            etree.fromstring(newstring, parser)
        except Exception:
            return None
        if escape_text_parts:
            newstring = EncodingXMLParser(newstring).parse()
        parsed_root = etree.fromstring(newstring, parser)
        return parsed_root if parsed_root.tag == xmltarget.tag else parsed_root[0]

    def _replace_xml_target(
        self,
        xmltarget: etree._Element,
        newtree: etree._Element,
    ) -> None:
        # Copy existing attributes
        for attribute, value in xmltarget.items():
            newtree.attrib[attribute] = value
        parent = xmltarget.getparent()
        # Parent can be none when operating on Unit only without a storage
        if parent is not None:
            parent.replace(xmltarget, newtree)
        # Update unit xmlelement if needed (plurals are inner elements here)
        if xmltarget == self.xmlelement:
            self.xmlelement = newtree

    def set_xml_text_cdata(self, target: str | None, xmltarget: etree._Element) -> None:
        if target is not None and "<![CDATA[" in target:
            newtree = self._parse_target_xml(
                xmltarget,
                target,
                escape_text_parts=True,
            )
            if newtree is not None and len(newtree) == 0:
                self._replace_xml_target(xmltarget, newtree)
                self._set_target_markup_mode(self.TARGET_MARKUP_CDATA)
                return

        # Remove possible old elements
        for x in xmltarget.iterchildren():
            xmltarget.remove(x)
        # Handle text only
        if target is None:
            xmltarget.text = None
        else:
            newtree = self._parse_target_xml(xmltarget, target)
            cdata_text = (
                self._extract_inner_xml(
                    etree.tostring(newtree, encoding="unicode", with_tail=False)
                )
                if newtree is not None
                else self.escape(target)
            )
            xmltarget.text = etree.CDATA(cdata_text)
        self._set_target_markup_mode(self.TARGET_MARKUP_CDATA)

    def set_xml_text_value(
        self,
        target: str | None,
        xmltarget: etree._Element,
        target_markup_mode: str | None = None,
    ) -> str:
        if target_markup_mode is None:
            target_markup_mode = self._get_target_markup_mode()
        if target is None:
            resulting_mode = (
                target_markup_mode
                if target_markup_mode
                in {self.TARGET_MARKUP_ESCAPED, self.TARGET_MARKUP_CDATA}
                else self.TARGET_MARKUP_PLAIN
            )
            if resulting_mode == self.TARGET_MARKUP_CDATA:
                self.set_xml_text_cdata(target, xmltarget)
            else:
                self.set_xml_text_plain(
                    target,
                    xmltarget,
                    target_markup_mode=resulting_mode,
                )
            return resulting_mode
        if target_markup_mode == self.TARGET_MARKUP_CDATA:
            self.set_xml_text_cdata(target, xmltarget)
            return target_markup_mode
        if target_markup_mode != self.TARGET_MARKUP_XML:
            self.set_xml_text_plain(
                target,
                xmltarget,
                target_markup_mode=target_markup_mode,
            )
            return target_markup_mode

        if "<" in target or "&" in target:
            # Try to handle it as legacy XML
            newtree = self._parse_target_xml(xmltarget, target)
            if newtree is None:
                self.set_xml_text_plain(
                    target,
                    xmltarget,
                    target_markup_mode=self.TARGET_MARKUP_PLAIN,
                )
                return self.TARGET_MARKUP_PLAIN
            self._replace_xml_target(xmltarget, newtree)
            resulting_mode = (
                self.TARGET_MARKUP_XML
                if len(newtree) != 0 or "<![CDATA[" in target
                else self.TARGET_MARKUP_PLAIN
            )
            self._set_target_markup_mode(resulting_mode)
            return resulting_mode
        self.set_xml_text_plain(
            target,
            xmltarget,
            target_markup_mode=self.TARGET_MARKUP_PLAIN,
        )
        return self.TARGET_MARKUP_PLAIN

    def gettargetlanguage(self):
        if self._store is None:
            return "en"
        return super().gettargetlanguage()

    @property
    def target(self):
        if self.xmlelement.tag != self.PLURAL_TAG:
            return self.get_xml_text_value(self.xmlelement)
        plurals = {
            entry.get("quantity"): self.get_xml_text_value(entry)
            for entry in self.xmlelement.iterchildren("item")
        }
        plural_tags = (
            self._store.get_plural_tags() if self._store else data.plural_tags["en"]
        )
        return multistring([plurals.get(tag, "") for tag in plural_tags])

    def fixup_tag(self, expected: str) -> None:
        # Fix the root tag if mismatching
        if self.xmlelement.tag != expected:
            old_id = self.getid()
            newelement = etree.Element(expected)
            parent = self.xmlelement.getparent()
            if parent is not None:
                parent.replace(self.xmlelement, newelement)
            self.xmlelement = newelement
            self.setid(old_id)

    @target.setter
    def target(self, target: TargetType) -> None:
        if self.hasplurals(self.source) or self.hasplurals(target):
            # Fix the root tag if mismatching
            self.fixup_tag(self.PLURAL_TAG)
            target_markup_mode = self._get_target_markup_mode()
            resulting_mode = (
                self.TARGET_MARKUP_PLAIN
                if target_markup_mode == self.TARGET_MARKUP_XML
                else target_markup_mode
            )

            plural_tags = self._store.get_plural_tags()  # ty:ignore[unresolved-attribute]

            # Sync plural_strings elements to plural_tags count.
            plural_target: str | multistring | list[str] = (
                "" if target is None else target
            )
            plural_strings = self.sync_plural_count(plural_target, plural_tags)

            # Rebuild plurals.
            for entry in self.xmlelement.iterchildren():
                self.xmlelement.remove(entry)

            self.xmlelement.text = "\n    "

            # Include additional plural for decimal numbers if not present. This is
            # enforced by Android lint but translate-toolkit currently does not support
            # editing this.
            locale = self._store.get_base_locale_code()  # ty:ignore[unresolved-attribute]
            if locale in data.DECIMAL_EXTRA_TAGS:
                for extra in data.DECIMAL_EXTRA_TAGS[locale]:
                    if extra not in plural_tags:
                        # Create copy here to avoid modifications to language.data
                        plural_tags = [*plural_tags, extra]
                        plural_strings.append(plural_strings[-1])

            # Include "other" as copy of "many" if "other" is not present. This avoids crashes
            # of Android builds with broken plurals handling.
            if "other" not in plural_tags and "many" in plural_tags:
                # Create copy here to avoid modifications to language.data
                plural_tags = [*plural_tags, "other"]
                plural_strings.append(plural_strings[-1])

            for plural_tag, plural_string in sorted(
                zip(plural_tags, plural_strings, strict=True),
                key=lambda item: data.cldr_plural_categories.index(item[0]),
            ):
                if not plural_string:
                    continue
                item = etree.Element("item")
                item.set("quantity", plural_tag)
                self.xmlelement.append(item)
                item_markup_mode = self.set_xml_text_value(
                    plural_string,
                    item,
                    target_markup_mode=target_markup_mode,
                )
                if (
                    target_markup_mode == self.TARGET_MARKUP_XML
                    and item_markup_mode == self.TARGET_MARKUP_XML
                ):
                    resulting_mode = self.TARGET_MARKUP_XML
            if target_markup_mode == self.TARGET_MARKUP_XML:
                self._set_target_markup_mode(resulting_mode)
            else:
                self._set_target_markup_mode(target_markup_mode)
        else:
            # Fix the root tag if mismatching
            self.fixup_tag(self.SINGULAR_TAG)

            self.set_xml_text_value(cast("str | None", target), self.xmlelement)

        self._rich_target = None
        self._target = target

    def getlanguageNode(self, lang=None, index=None):
        return self.xmlelement

    # Notes are handled as previous sibling comments.
    def addnote(self, text, origin=None, position="append") -> None:
        if origin in {"programmer", "developer", "source code", None}:
            self.xmlelement.addprevious(etree.Comment(text))
        else:
            super().addnote(text, origin=origin, position=position)

    def getnotes(self, origin=None):
        if origin in {"programmer", "developer", "source code", None}:
            comments = []
            if self.xmlelement is not None:
                prevSibling = self.xmlelement.getprevious()
                while (prevSibling is not None) and (prevSibling.tag is etree.Comment):
                    comments.insert(0, prevSibling.text)
                    prevSibling = prevSibling.getprevious()

            return "\n".join(comments)
        return super().getnotes(origin)

    def removenotes(self, origin=None) -> None:
        if (self.xmlelement is not None) and (self.xmlelement.getparent is not None):
            prevSibling = self.xmlelement.getprevious()
            while (prevSibling is not None) and (prevSibling.tag is etree.Comment):
                prevSibling.getparent().remove(prevSibling)
                prevSibling = self.xmlelement.getprevious()

        super().removenotes()

    def __str__(self) -> str:
        return etree.tostring(self.xmlelement, pretty_print=True, encoding="unicode")

    def __eq__(self, other):
        return str(self) == str(other)

    @staticmethod
    def hasplurals(thing):
        return isinstance(thing, (multistring, list))


class AndroidResourceFile(lisa.LISAfile):
    """Class representing an Android String resource file store."""

    UnitClass = AndroidResourceUnit
    Name = "Android String Resource"
    Mimetypes = ["application/xml"]
    Extensions = ["xml"]
    rootNode = "resources"
    bodyNode = "resources"
    XMLskeleton = """<?xml version="1.0" encoding="utf-8"?>
<resources></resources>"""
    XMLindent = {"indent": "    ", "leaves": {"string", "item"}, "trailing_eol": True}
    XMLuppercaseEncoding = False

    def initbody(self) -> None:
        """
        Initialises self.body so it never needs to be retrieved from the XML
        again.
        """
        self.namespace = self.document.getroot().nsmap.get(None, None)
        self.body = self.document.getroot()

    def parse(self, xml) -> None:
        """Populates this object from the given xml string."""
        if not hasattr(self, "filename"):
            self.filename = getattr(xml, "name", "")
        if hasattr(xml, "read"):
            xml.seek(0)
            posrc = xml.read()
            xml = posrc
        parser = get_safe_xml_parser(strip_cdata=False)
        self.document = etree.fromstring(xml, parser).getroottree()
        self._encoding = self.document.docinfo.encoding
        self.initbody()
        assert self.document.getroot().tag == self.namespaced(self.rootNode)

        for entry in self.document.getroot().iterchildren():
            term = self.UnitClass.createfromxmlElement(entry)
            if term is not None:
                self.addunit(term, new=False)

    def gettargetlanguage(self):
        target_lang = super().gettargetlanguage()

        # If targetlanguage isn't set, we try to extract it from the filename path (if any).
        if target_lang is None and hasattr(self, "filename") and self.filename:
            # Android standards expect resource files to be in a directory named "values[-<lang>[-r<region>]]".
            parent_dir = os.path.split(os.path.dirname(self.filename))[1]
            match = re.search(r"^values-(\w*)", parent_dir)
            if match is not None:
                target_lang = match.group(1)
            elif parent_dir == "values":
                # If the resource file is inside the "values" directory, then it is the default/source language.
                target_lang = self.sourcelanguage

            # Cache it
            self.settargetlanguage(target_lang)

        return target_lang

    def addunit(self, unit, new=True) -> None:
        """
        Adds unit to the document.

        In addition to the standard addunit, it also tries to move
        namespace definitions to the top <resources> element.
        """
        newns = {}
        do_cleanup = False
        if new:
            # Include any possible new namespaces
            newns = self.body.nsmap  # ty:ignore[unresolved-attribute]
            for ns in unit.xmlelement.nsmap:
                if ns not in newns:
                    do_cleanup = True
                    newns[ns] = unit.xmlelement.nsmap[ns]

            # Detect if unit is using XML entities
            hasentity = False
            for xmlelement in unit.xmlelement.iterdescendants():
                if xmlelement.tag is etree.Entity:
                    hasentity = True
                    break

            # Copy doctype to include possibly new entities
            if hasentity:
                cloned_doc = copy.deepcopy(unit._store.document)
                cloned_doc.getroot().clear()
                self.XMLdoctype = etree.tostring(
                    cloned_doc, xml_declaration=False, encoding="unicode"
                ).rsplit("\n", 1)[0]  # ty:ignore[invalid-argument-type]

        super().addunit(unit, new)
        # Move aliased namespaces to the <resources> tag
        # The top_nsmap was introduced in LXML 3.5.0
        if do_cleanup:
            etree.cleanup_namespaces(self.body, top_nsmap=newns)  # ty:ignore[invalid-argument-type]

    def removeunit(self, unit) -> None:
        unit.removenotes()
        super().removeunit(unit)


class CMPResourceUnit(AndroidResourceUnit):
    # Only \n and \t are escaped here (+ Unicode escape sequences), see
    # https://github.com/JetBrains/compose-multiplatform/blob/754dc253964e5aeab63581ed2a6a4ac376af1c32/gradle-plugins/compose/src/main/kotlin/org/jetbrains/compose/resources/PrepareComposeResources.kt#L270-L298
    ESCAPE_TRANSLATE = str.maketrans(
        {
            "\\": "\\\\",
            "\n": "\\n",
            "\t": "\\t",
        }
    )
    ESCAPE_ALL: ClassVar[bool] = False


class CMPResourceFile(AndroidResourceFile):
    UnitClass = CMPResourceUnit
    Name = "Compose Multiplatform Resources"


class MOKOResourceUnit(CMPResourceUnit):
    PLURAL_TAG = "plural"


class MOKOResourceFile(AndroidResourceFile):
    UnitClass = MOKOResourceUnit
    Name = "Mobile Kotlin Resources"



#
# Copyright 2016 Michal Čihař
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

r"""Class that manages YAML data files for translation."""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any

from ruamel.yaml import YAML, YAMLError
from ruamel.yaml.comments import CommentedMap, TaggedScalar
from ruamel.yaml.scalarstring import LiteralScalarString

from translate.lang.data import cldr_plural_categories
from translate.misc.multistring import multistring
from translate.storage import base

if TYPE_CHECKING:
    from collections.abc import Generator


class YAMLUnitId(base.UnitId):
    KEY_SEPARATOR = "->"
    INDEX_SEPARATOR = "->"

    def __str__(self) -> str:
        result = super().__str__()
        # Strip leading ->
        if result.startswith(self.KEY_SEPARATOR):
            return result[len(self.KEY_SEPARATOR) :]
        return result


class YAMLUnit(base.DictUnit):
    """A YAML entry."""

    IdClass = YAMLUnitId

    def __init__(self, source=None, **kwargs) -> None:
        # Ensure we have ID (for serialization)
        if source:
            self.source = source
            self._id = hex(hash(source))
        else:
            self._id = str(uuid.uuid4())
        super().__init__(source)

    @property
    def source(self):
        return self.target

    @source.setter
    def source(self, source) -> None:
        self.target = source

    def getid(self):
        return self._id

    def getlocations(self):
        return [self.getid()]

    def convert_target(self):
        return self.target

    def _get_original_value(self, output: dict[str, Any] | list[Any]) -> Any:
        """
        Get the original value from the YAML output structure to check its type.

        Args:
            output: The YAML output structure (dict or list) to navigate

        Returns:
            The original value if it exists, None if navigation fails or value doesn't exist

        """
        target = output
        parts = self.get_unitid().parts

        # Navigate to get the original value
        try:
            for part in parts:
                element, key = part
                if element in {"index", "key"}:
                    target = target[key]
        except (KeyError, IndexError, TypeError):
            return None
        return target

    def storevalue(
        self,
        output: dict[str, Any] | list[Any],
        value: Any,
        override_key: str | None = None,
        unset: bool = False,
    ) -> None:
        """Store value, preserving or converting to LiteralScalarString for multiline strings."""
        # Get the original value to check its type before it gets overwritten
        original_value = self._get_original_value(output)

        # Preserve or convert to LiteralScalarString for better readability
        # Always preserve LiteralScalarString if original was one, or
        # for new values or plain strings, use LiteralScalarString if multiline
        if (
            isinstance(value, str)
            and not unset
            and (
                isinstance(original_value, LiteralScalarString)
                or (
                    "\n" in value
                    and (original_value is None or type(original_value) is str)
                )
            )
        ):
            value = LiteralScalarString(value)
        # Otherwise keep the value as-is (e.g., DoubleQuotedScalarString stays quoted)

        # Call parent storevalue
        super().storevalue(output, value, override_key, unset)

    def storevalues(self, output: dict[str, Any] | list[Any]) -> None:
        self.storevalue(output, self.convert_target())


class YAMLFile(base.DictStore[YAMLUnit]):
    """A YAML file."""

    UnitClass = YAMLUnit

    def __init__(self, inputfile=None, **kwargs) -> None:
        """Construct a YAML file, optionally reading in from inputfile."""
        super().__init__(**kwargs)
        self.filename = ""
        self._original = self.get_root_node()
        self.dump_args = {
            "default_flow_style": False,
            "preserve_quotes": True,
        }
        if inputfile is not None:
            self.parse(inputfile)

    def get_root_node(self):
        """Returns root node for serialize."""
        return CommentedMap()

    @property
    def yaml(self):
        yaml = YAML()
        yaml.max_depth = 128
        for arg, value in self.dump_args.items():
            setattr(yaml, arg, value)
        return yaml

    def serialize(self, out) -> None:
        # Always start with valid root even if original file was empty
        if self._original is None:
            self._original = self.get_root_node()

        units = self.preprocess(self._original)
        self.serialize_units(units)
        self.yaml.dump(self._original, out)

    def _extract_comment_lines(self, tokens):
        """Extract comment text from YAML comment tokens."""
        comment_lines = []
        # Ensure tokens is a list
        if not isinstance(tokens, list):
            tokens = [tokens]

        for token in tokens:
            if hasattr(token, "value"):
                for line in token.value.split("\n"):
                    line = line.strip()
                    if line.startswith("#"):
                        comment_lines.append(line[1:].strip())
        return comment_lines

    def _get_key_comment(self, commented_map, key):
        """
        Extract the comment that appears before a key in a CommentedMap.

        Comments can appear in three places:
        1. Top-level comment for the first key (ca.comment[1])
        2. After the previous key's value (ca.items[prev_key][2])
        3. On separate lines before the key (ca.items[key][3])
        """
        if not isinstance(commented_map, CommentedMap) or not hasattr(
            commented_map, "ca"
        ):
            return None

        comment_lines = []
        keys = list(commented_map.keys())

        # Check for top-level comment if this is the first key
        if keys and keys[0] == key:
            ca_comment = getattr(commented_map.ca, "comment", None)
            if ca_comment and len(ca_comment) > 1 and ca_comment[1]:
                comment_lines.extend(self._extract_comment_lines(ca_comment[1]))

        # For non-first keys, check the previous key's end comment
        elif key in keys:
            key_index = keys.index(key)
            if key_index > 0 and hasattr(commented_map.ca, "items"):
                prev_key = keys[key_index - 1]
                prev_comment_info = commented_map.ca.items.get(prev_key)
                if (
                    prev_comment_info
                    and len(prev_comment_info) > 2
                    and prev_comment_info[2]
                ):
                    comment_lines.extend(
                        self._extract_comment_lines(prev_comment_info[2])
                    )

        # Check for comments on separate lines before this key
        if hasattr(commented_map.ca, "items"):
            comment_info = commented_map.ca.items.get(key)
            if comment_info and len(comment_info) > 3 and comment_info[3]:
                comment_lines.extend(self._extract_comment_lines(comment_info[3]))

        return "\n".join(comment_lines) if comment_lines else None

    def _parse_dict(self, data, prev):
        # Avoid using merged items, it is enough to have them once
        for k, v in data.non_merged_items():
            yield from self._flatten(v, prev.extend("key", k), parent_map=data, key=k)

    def _flatten(
        self, data, prev=None, parent_map=None, key=None
    ) -> Generator[tuple[base.UnitId, str, str | None]]:
        """
        Flatten YAML dictionary.

        Yields tuples of (unit_id, data, comment) where comment may be None.
        """
        if prev is None:
            prev = self.UnitClass.IdClass([])
        if isinstance(data, dict):
            yield from self._parse_dict(data, prev)
        elif isinstance(data, str):
            yield (prev, data, self._get_key_comment(parent_map, key))
        elif isinstance(data, (bool, int)):
            yield (prev, str(data), self._get_key_comment(parent_map, key))
        elif isinstance(data, list):
            for k, v in enumerate(data):
                yield from self._flatten(
                    v, prev.extend("index", k), parent_map=data, key=k
                )
        elif isinstance(data, TaggedScalar):
            yield (prev, data.value, self._get_key_comment(parent_map, key))
        elif data is None:
            pass
        else:
            raise ValueError(
                "We don't handle these values:\n"
                f"Type: {type(data)}\n"
                f"Data: {data}\n"
                f"Previous: {prev}"
            )

    @staticmethod
    def preprocess(data):
        """Preprocess hook for child formats."""
        return data

    def parse(self, input) -> None:  # ty:ignore[invalid-method-override]
        """Parse the given file or file source string."""
        if hasattr(input, "name"):
            self.filename = input.name
        elif not getattr(self, "filename", ""):
            self.filename = ""
        if hasattr(input, "read"):
            src = input.read()
            input.close()
            input = src
        if isinstance(input, bytes):
            input = input.decode("utf-8")
        try:
            self._original = self.yaml.load(input)
        except YAMLError as e:
            message = getattr(e, "problem", getattr(e, "message", str(e)))
            if hasattr(e, "problem_mark"):
                message += f" {e.problem_mark}"
            raise base.ParseError(message) from e

        content = self.preprocess(self._original)

        for k, data, comment in self._flatten(content):
            unit = self.UnitClass(data)
            unit.set_unitid(k)
            if comment:
                unit.addnote(comment, origin="developer")
            self.addunit(unit)

    def removeunit(self, unit) -> None:
        if self._original is not None:
            units = self.preprocess(self._original)
            unit.storevalue(units, None, unset=True)
        super().removeunit(unit)


class RubyYAMLUnit(YAMLUnit):
    def convert_target(self):
        if not isinstance(self.target, multistring):
            return self.target

        tags = self._store.get_plural_tags()  # ty:ignore[unresolved-attribute]

        # Sync plural_strings elements to plural_tags count.
        strings = self.sync_plural_count(self.target, tags)
        if any(strings):
            # Replace blank strings by None to distinguish not completed translations
            strings = [string or None for string in strings]

        return CommentedMap(zip(tags, strings, strict=True))


class RubyYAMLFile(YAMLFile):
    """Ruby YAML file, it has language code as first node."""

    UnitClass = RubyYAMLUnit

    def preprocess(self, data):
        if isinstance(data, CommentedMap) and len(data) == 1:
            lang = next(iter(data.keys()))
            # Handle blank values
            if data[lang] is None:
                data[lang] = CommentedMap()
            # Do not try to parse string only, CommentedMap is dict as well
            if isinstance(data[lang], dict):
                self.settargetlanguage(lang)
                return data[lang]
        return data

    def get_root_node(self):
        """Returns root node for serialize."""
        result = CommentedMap()
        result[self.targetlanguage or "en"] = CommentedMap()
        return result

    def _parse_dict(self, data, prev):
        # Does this look like a plural?
        tags = self.get_plural_tags()
        if data and all(x in cldr_plural_categories for x in data):
            # Ensure we have correct plurals ordering.
            values = [data[item] for item in tags if item in data]

            # Skip blank values (all plurals are None)
            if not all(value is None for value in values):
                # Use blank string instead of None here
                # Note: plurals don't have comments, so we pass None
                yield (prev, multistring([value or "" for value in values]), None)

            return

        # Handle normal dict
        yield from super()._parse_dict(data, prev)



#
# Copyright 2007-2008 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module contains all the common features for languages.

Supported features:

- language code (km, af)
- language name (Khmer, Afrikaans)
- Plurals

  - Number of plurals (nplurals)
  - Plural equation

- pofilter tests to ignore

Segmentation:

- characters
- words
- sentences

Punctuation:

- End of sentence
- Start of sentence
- Middle of sentence
- Quotes

  - single
  - double

- Valid characters
- Accelerator characters
- Special characters
- Direction (rtl or ltr)

TODOs and Ideas for possible features:

- Language-Team information
- Segmentation

  - phrases
"""

import logging
import re

from unicode_segmentation_rs import unicode_words

from translate.lang import data

logger = logging.getLogger(__name__)


class Common:
    """This class is the common parent class for all language classes."""

    code = ""
    """The ISO 639 language code, possibly with a country specifier or other
    modifier.

    Examples::

        km
        pt_BR
        sr_YU@Latn
    """

    fullname = ""
    """The full (English) name of this language.

    Dialect codes should have the form of:

    - Khmer
    - Portuguese (Brazil)
    - TODO: sr_YU@Latn?
    """

    nplurals = 0
    """The number of plural forms of this language.

    0 is not a valid value - it must be overridden.
    Any positive integer is valid (it should probably be between 1 and 6)

    .. seealso:: :mod:`translate.lang.data`
    """

    pluralequation = "0"
    """The plural equation for selection of plural forms.

    This is used for PO files to fill into the header.

    .. seealso::

       `Gettext manual <http://www.gnu.org/software/gettext/manual/html_node/gettext_150.html#Plural-forms>`_, :mod:`translate.lang.data`
    """
    # Don't change these defaults of nplurals or pluralequation willy-nilly:
    # some code probably depends on these for unrecognised languages

    mozilla_nplurals = 0
    mozilla_pluralequation = "0"
    """This of languages that has different plural formula in Mozilla than the
    standard one in Gettext."""

    listseperator = ", "
    """This string is used to separate lists of textual elements. Most
    languages probably can stick with the default comma, but Arabic and some
    Asian languages might want to override this."""

    specialchars = ""
    """Characters used by the language that might not be easy to input with
    common keyboard layouts"""

    commonpunc = ".,;:!?-@#$%^*_()[]{}/\\'`\"<>"
    """These punctuation marks are common in English and most languages that
    use latin script."""

    quotes = "‘’‛“”„‟′″‴‵‶‷‹›«»"
    """These are different quotation marks used by various languages."""

    invertedpunc = "¿¡"
    """Inverted punctuation sometimes used at the beginning of sentences in
    Spanish, Asturian, Galician, and Catalan."""

    rtlpunc = "،؟؛÷"
    """These punctuation marks are used by Arabic and Persian, for example."""

    CJKpunc = "。、，；！？「」『』【】"
    """These punctuation marks are used in certain circumstances with CJK
    languages."""

    indicpunc = "।॥॰"
    """These punctuation marks are used by several Indic languages."""

    ethiopicpunc = "።፤፣"
    """These punctuation marks are used by several Ethiopic languages."""

    miscpunc = "…±°¹²³·©®×£¥€"
    """The middle dot (·) is used by Greek and Georgian."""

    punctuation = f"{commonpunc}{quotes}{invertedpunc}{rtlpunc}{CJKpunc}{indicpunc}{ethiopicpunc}{miscpunc}"
    """We include many types of punctuation here, simply since this is only
    meant to determine if something is punctuation. Hopefully we catch some
    languages which might not be represented with modules. Most languages won't
    need to override this."""

    sentenceend = ".!?…։؟।。！？።\u06d4"
    """These marks can indicate a sentence end. Once again we try to account
    for many languages. Most languages won't need to override this."""

    # The following tries to account for a lot of things. For the best idea of
    # what works, see test_common.py. We try to ignore abbreviations, for
    # example, by checking that the following sentence doesn't start with lower
    # case or numbers.
    sentencere = re.compile(
        rf"""
        (?s)        # make . also match newlines
        .*?         # anything, but match non-greedy
        [{sentenceend}]        # the puntuation for sentence ending
        \s+         # the spacing after the puntuation
        (?=[^a-zа-џ\d])  # lookahead that next part starts with caps
        """,
        re.VERBOSE | re.UNICODE,
    )

    puncdict = {}
    """A dictionary of punctuation transformation rules that can be used by
    punctranslate()."""

    numbertuple = ()
    """A tuple of number transformation rules that can be used by
    numbertranslate()."""

    ignoretests = {}
    """Dictionary of tests to ignore in some or all checkers.

    Keys are checker names and values are list of names for the ignored tests
    in the checker. A special 'all' checker name can be used to tell that the
    tests must be ignored in all the checkers.

    Listed checkers to ignore tests on must be lowercase strings for the
    checker name, for example "mozilla" for MozillaChecker or "libreoffice" for
    LibreOfficeChecker."""

    checker = None
    """A language specific checker instance (see filters.checks).

    This doesn't need to be supplied, but will be used if it exists."""

    _languages = {}

    validaccel = None
    """Characters that can be used as accelerators (access keys) i.e. Alt+X
    where X is the accelerator.  These can include combining diacritics as
    long as they are accessible from the users keyboard in a single keystroke,
    but normally they would be at least precomposed characters. All characters,
    lower and upper, are included in the list."""

    validdoublewords = []
    """Some languages allow double words in certain cases.  This is a dictionary
    of such words."""

    def __new__(cls, code):
        """
        This returns the language class for the given code, following a
        singleton like approach (only one object per language).
        """
        code = code or ""
        # First see if a language object for this code already exists
        if code in cls._languages:
            return cls._languages[code]
        # No existing language. Let's build a new one and keep a copy
        language = cls._languages[code] = object.__new__(cls)

        language.code = code
        while code:
            langdata = data.get_language(code)
            if langdata:
                language.fullname, language.nplurals, language.pluralequation = langdata
                break
            code = data.simplercode(code)
        return language

    def __deepcopy__(self, memo={}):
        memo[id(self)] = self
        return self

    def __repr__(self) -> str:
        """
        Give a simple string representation without address information to
        be able to store it in text for comparison later.
        """
        detail = ""
        if self.code:
            detail = f"({self.code})"
        return f"<class 'translate.lang.common.Common{detail}'>"

    @classmethod
    def numbertranslate(cls, text):
        """
        Converts the numbers in a string according to the rules of the
        language.
        """
        if text:
            for latin_number, native_number in cls.numbertuple:
                text = text.replace(native_number, latin_number)
        return text

    @classmethod
    def punctranslate(cls, text):
        """
        Converts the punctuation in a string according to the rules of the
        language.
        """
        # TODO: look at po::escapeforpo() for performance idea
        if not text:
            return text
        ellipses_end = text.endswith("...")
        if ellipses_end:
            text = text[:-3]
        for source, target in cls.puncdict.items():
            text = text.replace(source, target)
        if ellipses_end:
            if "..." in cls.puncdict:
                text += cls.puncdict["..."]
            else:
                text += "..."
        # Let's account for cases where a punctuation symbol plus a space is
        # replaced, but the space won't exist at the end of the source message.
        # As a simple improvement for messages ending in ellipses (...), we
        # test that the last character is different from the second last
        # This is only relevant if the string has two characters or more
        if (f"{text[-1]} " in cls.puncdict) and (len(text) < 2 or text[-2] != text[-1]):
            text = text[:-1] + cls.puncdict[f"{text[-1]} "].rstrip()
        return text

    @classmethod
    def length_difference(cls, length):
        """
        Returns an estimate to a likely change in length relative to an
        English string of length length.
        """
        # This is just a rudimentary heuristic guessing that most translations
        # will be somewhat longer than the source language
        expansion_factor = 0
        code = cls.code
        while code:
            expansion_factor = data.expansion_factors.get(cls.code, 0)
            if expansion_factor:
                break
            code = data.simplercode(code)
        else:
            expansion_factor = 0.1  # default
        constant = max(5, int(40 * expansion_factor))
        # The default: return 5 + length/10
        return constant + int(expansion_factor * length)

    @classmethod
    def alter_length(cls, text):
        """
        Converts the given string by adding or removing characters as an
        estimation of translation length (with English assumed as source
        language).
        """

        def alter_it(text):
            length = len(text)
            if length > 9:
                extra = cls.length_difference(length)
                if extra > 0:
                    text = text[:extra].replace("\n", "") + text
                else:
                    text = text[-extra:]
            return text

        expanded = [alter_it(subtext) for subtext in text.split("\n\n")]
        return "\n\n".join(expanded)

    @classmethod
    def character_iter(cls, text):
        """Returns an iterator over the characters in text."""
        # We don't return more than one consecutive whitespace character
        prev = "A"
        for c in text:
            if c.isspace() and prev.isspace():
                continue
            prev = c
            if c not in cls.punctuation:
                yield c

    @classmethod
    def characters(cls, text):
        """Returns a list of characters in text."""
        return list(cls.character_iter(text))

    @classmethod
    def words(cls, text: str) -> list[str]:
        """Returns a list of words in text."""
        return unicode_words(text)

    @classmethod
    def sentence_iter(cls, text, strip=True):
        """Returns an iterator over the sentences in text."""
        lastmatch = 0
        text = text or ""
        for item in cls.sentencere.finditer(text):
            lastmatch = item.end()
            sentence = item.group()
            if strip:
                sentence = sentence.strip()
            if sentence:
                yield sentence
        remainder = text[lastmatch:]
        if strip:
            remainder = remainder.strip()
        if remainder:
            yield remainder

    @classmethod
    def sentences(cls, text, strip=True):
        """Returns a list of sentences in text."""
        return list(cls.sentence_iter(text, strip=strip))

    @classmethod
    def capsstart(cls, text):
        """Determines whether the text starts with a capital letter."""
        stripped = text.lstrip().lstrip(cls.punctuation)
        return stripped and stripped[0].isupper()

    @classmethod
    def numstart(cls, text):
        """Determines whether the text starts with a numeric value."""
        stripped = text.lstrip().lstrip(cls.punctuation)
        return stripped and stripped[0].isnumeric()



#
# Copyright 2004-2010 Zuza Software Foundation
#
# This file is part of translate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert template files (like .pot or template .xlf files) to translation
files, preserving existing translations.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/pot2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.misc.multistring import multistring
from translate.search import match
from translate.storage import catkeys, factory, poheader
from translate.tools import pretranslate


def convertpot(
    input_file,
    output_file,
    template_file,
    tm=None,
    min_similarity=75,
    fuzzymatching=True,
    maxlength=None,
    classes=None,
    classes_str=None,
    **kwargs,
) -> int:
    """Main conversion function."""
    input_store = factory.getobject(
        input_file, classes=classes, classes_str=classes_str
    )
    try:
        temp_store = factory.getobject(input_file, classes_str=classes_str)
    except Exception:
        # StringIO and other file like objects will be closed after parsing
        temp_store = None

    template_store = None
    if template_file is not None:
        template_store = factory.getobject(template_file, classes_str=classes_str)

    output_store = convert_stores(
        input_store,
        template_store,
        temp_store,
        tm,
        min_similarity,
        fuzzymatching,
        **kwargs,
    )
    convert.set_po_max_line_length(output_store, maxlength)
    output_store.serialize(output_file)

    return 1


def convert_stores(
    input_store,
    template_store,
    temp_store=None,
    tm=None,
    min_similarity=75,
    fuzzymatching=True,
    **kwargs,
):
    """
    Actual conversion function, works on stores not files, returns
    a properly initialized pretranslated output store, with structure
    based on input_store, metadata based on template_store, migrates
    old translations from template_store and pretranslating from TM.
    """
    if temp_store is None:
        temp_store = input_store

    # Create fuzzy matchers to be used by pretranslate.pretranslate_unit
    matchers = []

    _prepare_merge(input_store, temp_store, template_store)
    if fuzzymatching:
        if template_store:
            matcher = match.matcher(
                template_store,
                max_candidates=1,
                min_similarity=min_similarity,
                max_length=3000,
                usefuzzy=True,
            )
            matcher.addpercentage = False
            matchers.append(matcher)
        if tm:
            matcher = pretranslate.memory(
                tm, max_candidates=1, min_similarity=min_similarity, max_length=1000
            )
            matcher.addpercentage = False
            matchers.append(matcher)

    # initialize store
    _store_pre_merge(input_store, temp_store, template_store)

    # Do matching
    for input_unit in temp_store.units:
        if input_unit.istranslatable():
            input_unit = pretranslate.pretranslate_unit(
                input_unit,
                template_store,
                matchers,
                mark_reused=True,
                merge_on=input_store.merge_on,
            )
            _unit_post_merge(input_unit, input_store, temp_store, template_store)

    # finalize store
    _store_post_merge(input_store, temp_store, template_store)

    return temp_store


##dispatchers
def _prepare_merge(input_store, output_store, template_store, **kwargs) -> None:
    """Prepare stores & TM matchers before merging."""
    # Dispatch to format specific functions
    prepare_merge_hook = f"_prepare_merge_{input_store.__class__.__name__}"
    if prepare_merge_hook in globals():
        globals()[prepare_merge_hook](
            input_store, output_store, template_store, **kwargs
        )

    # Generate an index so we can search by source string and location later on
    input_store.makeindex()
    if template_store:
        template_store.makeindex()


def _store_pre_merge(input_store, output_store, template_store, **kwargs) -> None:
    """Initialize the new file with things like headers and metadata."""
    # formats that implement poheader interface are a special case
    if isinstance(input_store, poheader.poheader):
        _do_poheaders(input_store, output_store, template_store)
    elif isinstance(input_store, catkeys.CatkeysFile):
        # FIXME: also this should be a format specific hook
        if template_store is not None:
            output_store.header = template_store.header
        else:
            output_store.header = input_store.header

    # Dispatch to format specific functions
    store_pre_merge_hook = f"_store_pre_merge_{input_store.__class__.__name__}"
    if store_pre_merge_hook in globals():
        globals()[store_pre_merge_hook](
            input_store, output_store, template_store, **kwargs
        )


def _store_post_merge(input_store, output_store, template_store, **kwargs) -> None:
    """
    Close file after merging all translations, used for adding statistics,
    obsolete messages and similar wrapup tasks.
    """
    # Dispatch to format specific functions
    store_post_merge_hook = f"_store_post_merge_{input_store.__class__.__name__}"
    if store_post_merge_hook in globals():
        globals()[store_post_merge_hook](
            input_store, output_store, template_store, **kwargs
        )


def _unit_post_merge(
    input_unit, input_store, output_store, template_store, **kwargs
) -> None:
    """
    Handle any unit level cleanup and situations not handled by the merge()
    function.
    """
    # dispatch to format specific functions
    unit_post_merge_hook = f"_unit_post_merge_{input_unit.__class__.__name__}"
    if unit_post_merge_hook in globals():
        globals()[unit_post_merge_hook](
            input_unit, input_store, output_store, template_store, **kwargs
        )


## Format specific functions
def _unit_post_merge_pounit(
    input_unit, input_store, output_store, template_store
) -> None:
    """PO format specific plural string initialization logic."""
    # FIXME: do we want to do that for poxliff also?
    if input_unit.hasplural() and len(input_unit.target) == 0:
        # untranslated plural unit; Let's ensure that we have the correct
        # number of plural forms:
        nplurals, _plural = output_store.getheaderplural()
        if nplurals and nplurals.isdigit() and nplurals != "2":
            input_unit.target = multistring([""] * int(nplurals))


def _store_post_merge_pofile(input_store, output_store, template_store) -> None:
    """PO format specific: adds newly obsoleted messages to end of store."""
    # Let's take care of obsoleted messages
    if template_store:
        newlyobsoleted = []
        for unit in template_store.units:
            if unit.isheader() or unit.isblank():
                continue
            if unit.target and not (
                input_store.findid(unit.getid()) or hasattr(unit, "reused")
            ):
                # Not in .pot, make it obsolete
                unit.makeobsolete()
                newlyobsoleted.append(unit)
        for unit in newlyobsoleted:
            output_store.addunit(unit)


def _do_poheaders(input_store, output_store, template_store) -> None:
    """Adds initialized PO headers to output store."""
    # header values
    charset = "UTF-8"
    encoding = "8bit"
    project_id_version = None
    pot_creation_date = None
    po_revision_date = None
    last_translator = None
    language_team = None
    mime_version = None
    plural_forms = None
    kwargs = {}

    if template_store is not None and isinstance(template_store, poheader.poheader):
        templateheadervalues = template_store.parseheader()
        for key, value in templateheadervalues.items():
            if key == "Project-Id-Version":
                project_id_version = value
            elif key == "Last-Translator":
                last_translator = value
            elif key == "Language-Team":
                language_team = value
            elif key == "PO-Revision-Date":
                po_revision_date = value
            elif key in {"POT-Creation-Date", "MIME-Version"}:
                # don't know how to handle these keys, or ignoring them
                pass
            elif key == "Content-Type":
                kwargs[key] = value
            elif key == "Content-Transfer-Encoding":
                encoding = value
            elif key == "Plural-Forms":
                plural_forms = value
            else:
                kwargs[key] = value

    inputheadervalues = input_store.parseheader()
    for key, value in inputheadervalues.items():
        if key in {
            "Project-Id-Version",
            "Last-Translator",
            "Language-Team",
            "PO-Revision-Date",
            "Content-Type",
            "Content-Transfer-Encoding",
            "Plural-Forms",
        }:
            # want to carry these from the template so we ignore them
            pass
        elif key == "POT-Creation-Date":
            pot_creation_date = value
        elif key == "MIME-Version":
            mime_version = value
        else:
            kwargs[key] = value

    output_header = output_store.init_headers(
        charset=charset,
        encoding=encoding,
        project_id_version=project_id_version,
        pot_creation_date=pot_creation_date,
        po_revision_date=po_revision_date,
        last_translator=last_translator,
        language_team=language_team,
        mime_version=mime_version,
        plural_forms=plural_forms,
        **kwargs,
    )

    # Get the header comments and fuzziness state
    # override some values from input file
    if template_store is not None:
        template_header = template_store.header()
        if template_header is not None:
            if template_header.getnotes("translator"):
                output_header.addnote(
                    template_header.getnotes("translator"),
                    "translator",
                    position="replace",
                )
            output_header.markfuzzy(template_header.isfuzzy())


def main(argv=None) -> None:
    formats = {
        "pot": ("po", convertpot),
        ("pot", "po"): ("po", convertpot),
        "xlf": ("xlf", convertpot),
        ("xlf", "xlf"): ("xlf", convertpot),
        "xliff": ("xliff", convertpot),
        ("xliff", "xliff"): ("xliff", convertpot),
        "ts": ("ts", convertpot),
        "lang": ("lang", convertpot),
        ("lang", "lang"): ("lang", convertpot),
        ("ts", "ts"): ("ts", convertpot),
        "catkeys": ("catkeys", convertpot),
        ("catkeys", "catkeys"): ("catkeys", convertpot),
    }
    parser = convert.ConvertOptionParser(
        formats,
        usepots=True,
        usetemplates=True,
        allowmissingtemplate=True,
        description=__doc__,
    )

    parser.add_option(
        "",
        "--tm",
        dest="tm",
        default=None,
        help="The file to use as translation memory when fuzzy matching",
    )
    parser.passthrough.append("tm")

    defaultsimilarity = 75
    parser.add_option(
        "-s",
        "--similarity",
        dest="min_similarity",
        default=defaultsimilarity,
        type="float",
        help=f"The minimum similarity for inclusion (default: {defaultsimilarity}%)",
    )
    parser.passthrough.append("min_similarity")

    parser.add_option(
        "--nofuzzymatching",
        dest="fuzzymatching",
        action="store_false",
        default=True,
        help="Disable fuzzy matching",
    )
    parser.passthrough.append("fuzzymatching")

    parser.add_po_max_line_length_option()

    parser.run(argv)


if __name__ == "__main__":
    main()
